import os, inquirer
from . import PYONIR_SETUPS_DIRPATH, utilities


class PyonirSetup:
    create_app_msg = "Create a new Pyonir app"
    exit_cli_msg = "Exiting Pyonir cli"

    def create_new_app(self):
        """Creates new pyonir application directories"""
        self.decide_clear_dir()
        backend_dirpath = os.path.join(PYONIR_SETUPS_DIRPATH, 'backend')
        frontend_dirpath = os.path.join(PYONIR_SETUPS_DIRPATH, 'frontend')
        entry_filepath = os.path.join(PYONIR_SETUPS_DIRPATH, 'main.py')
        if not os.path.exists(self.app_path):
            utilities.copy_assets(backend_dirpath, os.path.join(self.app_path, 'backend'), False)
            utilities.copy_assets(entry_filepath, os.path.join(self.app_path, 'main.py'), False)
        if self.app_use_frontend:
            utilities.copy_assets(frontend_dirpath, os.path.join(self.app_path, 'frontend'), False)
            pass

    def __init__(self):
        self.app_path = None
        self.frontend_tool = None
        self.app_use_frontend = None
        self.app_name: str = ""
        self.user_dir: str = os.getcwd()
        self.action: str = ""
        self.action_step: str = ""
        self.summary: str = ""
        self.intro()

    def intro(self):
        print("Welcome to the Pyonir Web framework!")
        # Step 1: Choose an action
        self.action = inquirer.prompt([
            inquirer.List('action',
                          message="What would you like to do?",
                          choices=[self.create_app_msg, self.exit_cli_msg],
                          ),
        ])['action']

        # Process user selection
        if self.action == self.create_app_msg:
            self.decide_project_name()
            self.decide_use_frontend()
        elif self.action == self.exit_cli_msg:
            print(self.summary)
            exit(0)

    def outro(self):
        self.action_step = inquirer.prompt([
            inquirer.List('action_step',
                          message="What would you like to configure next?",
                          choices=['Go back to main menu', 'Run setup'],
                          ),
        ])['action_step']
        self.summary = f'''
Project {self.app_name} created!
- path: {self.app_path}
- use frontend: {self.app_use_frontend}
    - install frontend: {self.frontend_tool}
        '''
        if self.action_step == 'Go back to main menu':
            self.intro()
        else:
            print(self.summary)
            self.create_new_app()
            print("Exiting the app.")
            exit(0)

    def decide_project_name(self):
        # print("You chose Task 1!")
        # Additional options or tasks for Task 1
        # Choose an App Name
        self.app_name = inquirer.prompt([
            inquirer.Text('app_name',
                          message="What is your project named?"
                          ),
        ])['app_name']

        self.app_path = os.path.join(self.user_dir, self.app_name)

    def decide_clear_dir(self):
        if os.path.exists(self.app_path):
            override = inquirer.prompt([
                        inquirer.Text('override',
                      message="Pyonir app already exists. Would you like to start with a clean directory?"
                      ),
                    ])['override']
            if override:
                print('delete this directory', self.app_path)

    def decide_use_frontend(self):
        # Choose optional Frontend
        app_use_frontend = inquirer.prompt([
            inquirer.Text('app_use_frontend',
                          message="Would you like to use a frontend? (y=Yes or n=No)"
                          ),
        ])['app_use_frontend']
        self.app_use_frontend = True if app_use_frontend.lower() == 'y' else False
        if self.app_use_frontend:
            self.decide_frontend_tool()

        self.outro()

    def decide_frontend_tool(self):
        self.frontend_tool = inquirer.prompt([
            inquirer.List('use_frontend_tool',
                          message="Which frontend tool would you like to install?",
                          choices=["Deno", "Bun", "Vite"],
                          ),
        ])['use_frontend_tool']


PyonirSetup()



from pyonir.types import PyonirRequest


async def demo_items(req: PyonirRequest, sample_id: int):
    """Home route handler"""
    return f"Main app ITEMS route {sample_id}!"


# Define routes
routes = [
    ['/items', demo_items, ["GET"]],
    ['/items/{sample_id:int}', demo_items, ["GET"]],
]

# Define an endpoint
endpoints = [
    ('/api/demo', routes)
]



from jinja2 import nodes
from jinja2.ext import Extension


def init():
    return MinifyTemplate

class MinifyTemplate(Extension):

    tags = {'minify'}
    def __init__(self, environment):
        super(MinifyTemplate, self).__init__(environment)

    def parse(self, parser):
        lineno = next(parser.stream).lineno
        body = parser.parse_statements(['name:endminify'], drop_needle=True)
        return nodes.CallBlock(self.call_method('_minify'), [], [], body).set_lineno(lineno)

    @staticmethod
    def _minify(caller):
        return caller().strip()
        # try:
        #     return html_minify(caller())
        # except:
        #     return caller().strip()
        # return minify(
        #     caller(),
        #     remove_comments=True,
        #     reduce_boolean_attributes=True,
        #     remove_optional_attribute_quotes=False )



import asyncio
import os, typing, json, inspect

from starlette.websockets import WebSocket, WebSocketState, WebSocketDisconnect

from pyonir import ASSETS_ROUTE, UPLOADS_ROUTE, PAGINATE_LIMIT
from pyonir.parser import ParselyPage
from pyonir.types import IApp, PyonirRequest, PyonirServer, PyonirHooks
from pyonir.utilities import Collection, create_file, get_attr

TEXT_RES = 'text/html'
JSON_RES = 'application/json'
EVENT_RES = 'text/event-stream'

ConnClients = {}


async def pyonir_ws_handler(websocket: WebSocket):
    """ws connection handler"""
    from pyonir.utilities import generate_id

    async def get_data(ws: WebSocket):
        assert ws.application_state == WebSocketState.CONNECTED and ws.client_state == WebSocketState.CONNECTED
        wsdata = await ws.receive()

        if wsdata.get('text'):
            wsdata = wsdata['text']
            swsdata = json.loads(wsdata)
            swsdata['value'] = swsdata.get('value')
            wsdata = json.dumps(swsdata)
        elif wsdata.get('bytes'):
            wsdata = wsdata['bytes'].decode('utf-8')

        return wsdata

    async def broadcast(message: str, ws_id: str = None):
        for id, ws in ConnClients.items():
            if active_id == id and hasattr(ws, 'send_text'): continue
            await ws.send_text(message)

    async def on_disconnect(websocket: WebSocket):
        del ConnClients[active_id]
        client_data.update({"action": "ON_DISCONNECTED", "id": active_id})
        await broadcast(json.dumps(client_data))

    async def on_connect(websocket: WebSocket):
        client_data.update({"action": "ON_CONNECTED", "id": active_id})
        await websocket.send_text(json.dumps(client_data))

    active_id = generate_id()
    client_data = {}
    await websocket.accept()  # Accept the WebSocket connection
    print("WebSocket connection established!")
    ConnClients[active_id] = websocket
    await on_connect(websocket)
    try:
        while websocket.client_state == WebSocketState.CONNECTED:
            # Wait for a message from the client
            data = await get_data(websocket)
            print(f"Received from client: {data}")
            # Respond to the client
            await broadcast(data)
        await on_disconnect(data)
    except Exception as e:
        del ConnClients[active_id]
        print(f"WebSocket connection closed: {active_id}")


async def pyonir_sse_handler(request: PyonirRequest) -> typing.AsyncGenerator:
    """Handles sse web request by pyonir"""
    from pyonir.utilities import generate_id
    request.type = EVENT_RES  # assign the appropriate streaming headers
    # set sse client
    event = request.query_params.get('event')
    retry = request.query_params.get('retry', 1000)
    close_id = request.query_params.get('close')
    interval = 1  # time between events
    client_id = request.query_params.get('id', request.headers.get('user-agent') + f"{generate_id()}")
    if close_id and ConnClients.get(close_id):
        del ConnClients[close_id]
        return
    last_client = ConnClients.get(client_id, {
        "retry": retry,
        "event": event,
        "id": client_id,
        "data": {
            "time": 0
        },
    })
    # register new client
    if not ConnClients.get(client_id):
        ConnClients[client_id] = last_client

    while True:
        last_client["data"]["time"] = last_client["data"]["time"] + 1
        is_disconnected = await request.server_request.is_disconnected()
        if is_disconnected or close_id:
            del ConnClients[client_id]
            break
        await asyncio.sleep(interval)  # Wait for 5 seconds before sending the next message
        res = process_sse(last_client)
        yield res


def pyonir_file_delete(request: PyonirRequest):
    """Deletes a file located in the uploads directory"""
    from pyonir import Site
    from pyonir.parser import ParselyMedia
    redirect = request.query_params.get('redirect') + '?success=true'
    doc = os.path.join(Site.contents_dirpath, request.query_params.get('file'))
    img = ParselyMedia(doc, Site.files_ctx)
    if img.file_exists:
        os.remove(doc)
        for _, timg in img.thumbnails.items():
            os.remove(timg.abspath)
    return {"redirect": redirect}


async def pyonir_file_upload(request: PyonirRequest):
    """File upload handler"""
    from pyonir.parser import ParselyMedia
    from pyonir import Site
    folder_name = request.form.get('foldername')
    uploads = []
    for doc in request.files:
        parselyMedia = await ParselyMedia.save_upload(doc, os.path.join(Site.uploads_dirpath, folder_name))
        parselyMedia.resize()
        uploads.append(parselyMedia.data)
    return uploads


def pyonir_index(request: PyonirRequest):
    """Catch all routes for all web request"""
    return f"Pyonir default controller."


async def apply_plugin_resolvers(page: ParselyPage, request: PyonirRequest):
    """Resolves api definitions for plugins"""
    from pyonir import Site, utilities, CONTROLLERS_DIRNAME
    resolver_path = page.data.get('@resolver', {}).get(request.method)
    if not resolver_path: return
    pkg, meth_name = resolver_path.split('.', 1)
    if pkg == 'pyonir':
        resolver = get_attr(Site.server.resolvers, meth_name)
    else:
        pkg_path = os.path.join(Site.backend_dirpath, CONTROLLERS_DIRNAME, pkg, '__init__.py')
        module, resolver = utilities.get_module(pkg_path, meth_name)

    if not resolver: return
    is_async = inspect.iscoroutinefunction(resolver)
    rdata = await resolver(**request.args) if is_async else resolver(**request.args) if callable(resolver) else resolver
    return rdata if inspect.iscoroutine(rdata) or inspect.isasyncgen(rdata) else page.output_json(rdata)


async def process_request_data(request: PyonirRequest):
    """Get form data and file contents from request"""
    from pyonir import Site
    # from pyonir.parser import ParselyMedia
    from .utilities import secure_upload_filename
    try:
        try:
            ajson = await request.server_request.json()
            if isinstance(ajson, str): ajson = json.loads(ajson)
            request.form.update(ajson)
        except Exception as ee:
            # multipart/form-data
            form = await request.server_request.form()
            files = []
            for name, content in form.multi_items():
                if name == 'files':
                    # filedata = await content.read()
                    mediaFile = (secure_upload_filename(content.filename), content, Site.uploads_dirpath)
                    request.files.append(mediaFile)
                else:
                    request.form[name] = content
    except Exception as e:
        raise


def setup_starlette_server(iapp: IApp) -> PyonirServer:
    """Setup Starlette web server"""

    from starlette_wtf import CSRFProtectMiddleware
    from starlette_session import SessionMiddleware
    from starlette.middleware.trustedhost import TrustedHostMiddleware
    from starlette.middleware.gzip import GZipMiddleware
    from starlette.responses import Response, StreamingResponse
    # from starlette.requests import Request
    from starlette.routing import Router, Route

    def render_response(value, media_type):
        if media_type == EVENT_RES:
            return StreamingResponse(content=value, media_type=media_type)
        return Response(content=value, media_type=media_type)

    def redirect(url, code=302):
        from starlette.responses import RedirectResponse
        res = RedirectResponse(url=url.strip(), status_code=code)
        return res

    def get_staticroute(assets_dir):
        from starlette.staticfiles import StaticFiles
        return StaticFiles(directory=assets_dir)

    allowed_hosts = ['localhost', '*.localhost']
    secret_sauce = getattr(iapp.configs.app, 'secret_sauce', iapp.SECRET_SAUCE)
    session_key = getattr(iapp.configs.app, 'session_key', iapp.SESSION_KEY)
    star_app = PyonirServer()  # inherits from Starlette
    star_app.add_middleware(TrustedHostMiddleware)
    # star_app.add_middleware(GZipMiddleware, minimum_size=500)
    star_app.add_middleware(SessionMiddleware,
                            https_only=True,
                            secret_key=secret_sauce,
                            cookie_name=session_key
                            )
    star_app.add_middleware(CSRFProtectMiddleware, csrf_secret=secret_sauce)

    # Interface properties required for pyonir to process web request
    setattr(star_app, 'response_renderer', render_response)
    setattr(star_app, 'serve_redirect', redirect)
    setattr(star_app, 'create_endpoint', Router)
    setattr(star_app, 'create_route', Route)
    setattr(star_app, 'serve_static', get_staticroute)

    return star_app


def url_for(name, attr='path'):
    """returns application route value based on named class function name"""
    from pyonir import Site
    if not Site: return None
    urlmap = Site.server.url_map
    return urlmap.get(name, {}).get(attr)


def init_endpoints(endpoints: 'Endpoints'):
    for endpoint, routes in endpoints:
        for path, func, methods, *opts in routes:
            args = opts[0] if opts else {}
            kwargs = {'models': {}}
            for k, v in args.items():
                if k in ('ws', 'sse', 'static_path'):
                    kwargs[k] = v
                kwargs['models'].update({k: v})
            r = add_route(f'{endpoint}{path}', func, methods, **kwargs)
            pass


def init_parsely_endpoints(app: IApp):
    for r, static_abspath in ((ASSETS_ROUTE, app.theme_static_dirpath), (UPLOADS_ROUTE, app.uploads_dirpath)):
        if not os.path.exists(static_abspath): continue
        add_route(r, None, static_path=static_abspath)
    add_route("/syssse", pyonir_sse_handler, sse=True)
    add_route("/sysws", pyonir_ws_handler, ws=True)
    add_route("/", pyonir_index, methods='*')
    add_route("/{path:path}", pyonir_index, methods='*')


def get_params(url):
    import urllib
    args = {params.split('=')[0]: urllib.parse.unquote(params.split("=").pop()) for params in
            url.split('&') if params != ''}
    if args.get('model'): del args['model']
    return args


def process_sse(data: dict) -> str:
    """Formats a string and an event name in order to follow the event stream convention.
    'event: Jackson 5\\ndata: {"abc": 123}\\n\\n'
    """
    sse_payload = ""
    for key, val in data.items():
        val = json.dumps(val) if key == 'data' else val
        sse_payload += f"{key}: {val}\n"
    return sse_payload + "\n"


def serve_favicon(app):
    from starlette.responses import FileResponse
    return FileResponse(os.path.join(app.theme_static_dirpath,'favicon.ico'), 200)


def add_route(path: str,
              dec_func: typing.Callable,
              methods=None,
              models: dict = None,
              auth: bool = None,
              ws: bool = None,
              sse: bool = None,
              static_path: str = None) -> typing.Callable | None:
    """Route decorator"""
    from pyonir import Site

    def is_async(func):
        return inspect.isasyncgenfunction(func) or inspect.iscoroutinefunction(func)

    is_async = inspect.iscoroutinefunction(dec_func) if dec_func else False
    is_asyncgen = inspect.isasyncgenfunction(dec_func) if dec_func else False
    list_of_args = list(inspect.signature(dec_func).parameters.keys()) if dec_func else None
    if methods == '*':
        methods = ['GET', 'POST', 'PUT', 'PATCH', 'DELETE']
    if methods is None:
        methods = ['GET']

    if static_path:
        from starlette.staticfiles import StaticFiles
        static_route = Site.server.serve_static(static_path)
        Site.server.mount(path, static_route)
        Site.server.url_map[path.lstrip('/')
        .replace('/', '_')] = {'path': path, 'dir': static_path, 'exists': os.path.exists(static_path)}
        return

    route_name = dec_func.__name__
    docs = dec_func.__doc__
    route_path = path.split('/{')[0]
    name = route_name
    endpoint_route = path.split('/', 1)[0]
    is_pyonir_default = dec_func.__name__ == 'pyonir_index'
    req_models = Site.server.url_map.get(route_name, {}).get('models') or {}
    if models:
        for req_param, req_model in models.items():
            req_models[req_param] = req_model.__name__
            if hasattr(Site.server, 'models'):
                Site.server.models.update({req_model.__name__, req_model})
            else:
                Site.server.models = {req_model.__name__, req_model}
    new_route = {
        "doc": docs,
        "endpoint": endpoint_route,
        "route": path,  # has regex pattern
        "path": route_path,
        "methods": methods,
        "models": models or req_models,
        "name": name,
        "auth": auth,
        "sse": sse,
        "ws": ws,
        "async": is_async,
        "async_gen": is_asyncgen,
    }
    # Add route path into categories
    Site.server.endpoints.append(f"{endpoint_route}{route_path}")
    Site.server.url_map[name] = new_route
    if sse:
        Site.server.sse_routes.append(f"{endpoint_route}{route_path}")
    if auth:
        Site.server.auth_routes.append(f"{endpoint_route}{route_path}")
    if ws:
        Site.server.ws_routes.append(f"{endpoint_route}{route_path}")
        return Site.server.add_websocket_route(path, dec_func, dec_func.__name__)

    async def dec_wrapper(star_req):
        from pyonir.parser import ParselyPage
        if star_req.url.path == '/favicon.ico': return serve_favicon(Site)
        # Resolve page file route
        app_ctx, req_filepath = resolve_path_to_file(star_req.url.path)
        pyonir_request = build_request(star_req)
        await process_request_data(pyonir_request)
        if pyonir_request.type == TEXT_RES and not os.path.exists(app_ctx.frontend_dirpath):
            pyonir_request.type = JSON_RES
        route_models = app_ctx.server.url_map.get(dec_func.__name__, {}).get('models')

        args = pyonir_request.request_model(list_of_args, route_models)
        pyonir_request.args = args

        # Update template global
        app_ctx.TemplateParser.globals['request'] = pyonir_request

        # Resolve page from request
        req_file = ParselyPage(req_filepath, app_ctx.files_ctx)
        pyonir_request.file = req_file
        # Resolve route decorator methods
        pyonir_request.server_response = await dec_func(**args) if is_async else dec_func(**args)
        await app_ctx.run_plugins(PyonirHooks.ON_REQUEST, pyonir_request)
        # Finalize response output
        pyonir_request.status_code = pyonir_request.derive_status_code(is_pyonir_default)
        if pyonir_request.path not in app_ctx.server.sse_routes + app_ctx.server.ws_routes:
            pyonir_request.server_response = await req_file.process_response(pyonir_request)

        if pyonir_request.redirect:
            return Site.server.serve_redirect(pyonir_request.redirect, 303)

        return build_response(pyonir_request)

    Site.server.add_route(path, dec_wrapper, methods=methods)


def build_request(TRequest, code: int = 444) -> PyonirRequest:
    """Transforms generic server request type into Pyonir request object"""
    from pyonir import Site
    raw_path = TRequest.url.path
    method = TRequest.method
    path = TRequest.url.path
    path_params = TRequest.path_params
    url = f"{path}"
    slug = path.lstrip('/').rstrip('/')
    query_params = get_params(TRequest.url.query)
    parts = slug.split('/') if slug else []
    limit = Collection.get_attr(query_params, 'limit', PAGINATE_LIMIT)
    model = query_params.get('model')
    is_home = (path == '')
    form = {}
    files = []
    ip = TRequest.client.host
    host = str(TRequest.base_url).rstrip('/')
    protocol = TRequest.scope.get('type') + "://"
    headers = PyonirRequest.process_header(TRequest.headers)
    browser = headers.get('user-agent', '').split('/').pop(0) if headers else "UnknownAgent"
    if slug.startswith('api'): headers['accept'] = JSON_RES
    res_type = headers.get('accept')
    status_code = code
    auth = None
    use_endpoints = TRequest.url.path in Site.server.endpoints
    args = {}
    return PyonirRequest(raw_path, method, path, path_params, url, slug, query_params, parts, limit, model, is_home,
                         form, files, ip, host, protocol, headers, args, browser, res_type, status_code, auth,
                         use_endpoints=use_endpoints,
                         server_request=TRequest, file=None)


def build_response(request: PyonirRequest):
    """Create web response for web server"""
    from datetime import datetime, timedelta
    from pyonir import Site
    try:
        ishtml = request.type == TEXT_RES
        expires = datetime.utcnow() + timedelta(days=7)
        expires = expires.strftime("%a, %d %b %Y %H:%M:%S GMT")
        response = Site.server.response_renderer(request.server_response, media_type=request.type)
        if ishtml: response.headers['Expires'] = expires
        response.headers['Cache-Control'] = "no-cache" if not ishtml else "public, max-age=0"
        response.headers['Server'] = "Pyonir Web Framework"
        response.status_code = request.status_code
        return response
    except Exception as e:
        raise


def get_route_ctx(app: IApp, path_str: str) -> tuple:
    """Gets the routing context from web request"""
    path_str = path_str.replace('api/', '')
    req_trgt = path_str.split('/').pop()
    res = None
    for ctxp in app.request_paths:
        app_ctx, route, pths, prunes = ctxp
        ishome = route == path_str
        prunes = [p for p in prunes if p != req_trgt or ishome]
        if not path_str.startswith(route.lstrip('/')): continue
        if ishome: route = ''
        res = (app if isinstance(app_ctx, str) else app_ctx), route, pths, prunes
    return res


def resolve_path_to_file(path_str: str, skip_vanity: bool = False) -> typing.Tuple[IApp, str]:
    """Returns application name and requested file path based on web path"""
    from pyonir import Site, EXTENSIONS
    app_ctx, ctx_route, ctx_paths, prunes = get_route_ctx(Site, path_str)
    reqst_path = [p for p in path_str.split('/') if p not in prunes]
    path_result = None
    for p in ctx_paths:
        is_cat = os.path.join(p, *reqst_path, 'index.md')
        is_page = os.path.join(p, *reqst_path) + EXTENSIONS['file']
        for pp in (is_cat, is_page):
            if not os.path.exists(pp): continue
            path_result = pp
            break
        if path_result: break

    # if path_result is None and not skip_vanity:
    #     path_result = Request.vanity_url('/'+path_str)
    return app_ctx, path_result


def generate_nginx_conf(app: IApp):
    """Generates a NGINX conf file based on App configurations"""
    nginxconf = app.TemplateParser.get_template("nginx.jinja.conf") \
        .render(
        app_name=app.configs.app.name,
        app_name_id=app.configs.app.name.replace(' ', '_').lower(),
        domain=app.domain,
        is_secure=app.is_secure,
        serve_static=True,
        site_dirpath=app.app_dirpath,
        site_logs_dirpath=app.site_logs_dirpath,
        app_socket_filepath=app.app_socket_filepath,
        site_assets_route='/public',
        site_theme_assets_dirpath=app.theme_assets_dirpath,
        site_uploads_route='/uploads',
        site_uploads_dirpath=app.uploads_dirpath,
        site_ssg_dirpath=app.ssg_dirpath,
        custom_nginx_locations=Collection.get_attr(app.configs, 'nginx_locations')
    )

    create_file(app.app_nginx_conf_filepath, nginxconf, False)


def start_uvicorn_server(app: IApp, endpoints: 'Endpoints'):
    """Starts the webserver"""
    import uvicorn

    # """Uvicorn web server configurations"""
    from pyonir import PYONIR_SSL_KEY, PYONIR_SSL_CRT
    uvicorn_options = {
        "port": app.port,
        "host": app.host
    }
    if app.is_secure:
        uvicorn_options["ssl_keyfile"] = PYONIR_SSL_KEY
        uvicorn_options["ssl_certfile"] = PYONIR_SSL_CRT
    if not app.is_dev:
        uvicorn_options['uds'] = app.app_socket_filename

    # Initialize routers
    init_endpoints(endpoints)
    init_parsely_endpoints(app)
    # init_controllers(self)
    print(f"/************** ASGI APP SERVER RUNNING on {'http' if app.is_dev else 'sock'} ****************/")
    print(f"\
    \n\t- Sys User: {app.SYS_USER}\
    \n\t- App Version: {app.VERSION}\
    \n\t- App env: {'DEV' if app.is_dev else 'PROD'}\
    \n\t- App name: {app.name}\
    \n\t- App domain: {app.origin}\
    \n\t- App host: {app.host}\
    \n\t- App port: {app.port}\
    \n\t- App sock: {app.app_socket_filepath}\
    \n\t- App Server: Uvicorn \
    \n\t- NGINX config: {app.app_nginx_conf_filepath} \
    \n\t- System Version: {app.SYS_VERSION}")
    uvicorn.run(app.server, **uvicorn_options)



import os
from pyonir.types import IPlugin, PyonirRequest, IApp
from pyonir.parser import Schema
from pyonir.utilities import parse_all


class Navigation(IPlugin):
    """Assembles a map of navigation menus based on file configurations"""
    name = 'Navigation Plugin'

    def __init__(self, app: 'IApp'):
        menu_schema_dirpath = os.path.join(os.path.dirname(__file__), 'backend/contents/schemas')
        self.schemas = parse_all(menu_schema_dirpath, app.files_ctx, Schema)
        self.menus = {}
        pass

    async def on_request(self, request: PyonirRequest, app: IApp):
        from pyonir.utilities import Collection, allFiles
        if app is None: return None
        assert hasattr(app, 'pages_dirpath'), "Get menus 'app' parameter does not have a pages dirpath property"
        refresh_nav = bool(request.query_params.get('rnav'))
        curr_nav = app.TemplateParser.globals.get('navigation')
        if curr_nav and not refresh_nav: return False
        pages = {}
        subpages = {}
        active_page = request.path

        file_list = allFiles(app.pages_dirpath, app_ctx=app.files_ctx)  # return files using menu schema model

        for pg in file_list:
            page = self.schemas.menu.map_input_to_model(pg)
            has_menu = hasattr(page,'menu') or hasattr(page,'menu_parent')
            if page.status == 'hidden' or not page.url or (not has_menu): continue
            page.active = active_page == page.url
            if page.menu:
                pages[page.url] = page
            elif page.menu_parent:
                _ref = subpages.get(page.menu_parent)
                if not _ref:
                    subpages[page.menu_parent] = [page]
                else:
                    _ref.append(page)

        if subpages:
            for k, m in subpages.items():
                pmenu = pages.get(k)
                if not pmenu: continue
                pmenu.update({"sub_menu": m})

        res = Collection(pages.values(), sortBy='order').groupedBy('menu')
        self.menus[app.name] = res
        app.TemplateParser.globals['navigation'] = res



import os, pytz, re, json
from datetime import datetime
from typing import Generator

from .types import PyonirRequest, PyonirHooks
from .utilities import get_attr, allFiles, tupleconverter, deserialize_datestr, Collection, create_file, \
    remove_html_tags, dict_to_class

ALLOWED_CONTENT_EXTENSIONS = ('prs', 'md', 'json', 'yaml')
IGNORE_FILES = ('.vscode', '.vs', '.DS_Store', '__pycache__', '.git', '.', '_', '<', '>', '(', ')', '$', '!', '._')

REG_ILN_LIST = r'([-$@\s*=\w.]+)(\:-)(.*)'
REG_MAP_LST = r'(^[-$@\s*=\w.]+)(\:[`:`-]?)(.*)'
REG_METH_ARGS = r"\(([^)]*)\)"
DICT_DELIM = ":"
LST_DLM = ":-"
STR_DLM = ":`"
ILN_DCT_DLM = ":: "
BLOCK_PREFIX_STR = "==="
LOOKUP_EMBED_PREFIX = '$'
LOOKUP_FILE_PREFIX = '$file'
LOOKUP_CONTENT_PREFIX = '$content'
LOOKUP_DIR_PREFIX = '$dir'
FILTER_KEY = '@filter'
DEFAULT_IGNORE_FIELDS = ()
EmbeddedTypes = dict()
# Schemas configs
PROTECTED_FIELD_PREFIX = '@'
# Private fields should can be read and written but should not be exposed
PRIVATE_FIELDS = ('password', 'ssn', 'auth_token', 'id', 'uid', 'privateKey')
# Protected fields are allowed to be read but are not allowed to be changed directly
PROTECTED_FIELDS = (
                       'created_on', 'modified_on', 'modified_by', 'last_modified_by',
                       'date_created', 'date_modified', 'raw', 'provider_model') + PRIVATE_FIELDS

# Image configs
IMG_FILENAME_DELIM = '|'  # delimits the file name and description
IMAGE_FORMATS = (
    'JPEG',  # .jpg, .jpeg
    'JPG',  # .jpg, .jpeg
    'PNG',  # .png
    'GIF',  # .gif
    'BMP',  # .bmp
    'TIFF',  # .tiff, .tif
    'ICO',  # .ico
    'PPM',  # .ppm
    'PGM',  # .pgm
    'PBM',  # .pbm
    'WebP',  # .webp
    'DDS',  # .dds
    'TGA',  # .tga
    'XBM',  # .xbm
    'PCX'  # .pcx
)


def parse_markdown(content, kwargs):
    from mistletoe import markdown
    from mistletoe.html_attributes_renderer import HTMLAttributesRenderer
    return markdown(content, renderer=HTMLAttributesRenderer)


class Parsely:
    """Parsely is a static file parser"""

    # Extensions = {}  # Global extensions map that modify return values for parsely file data
    Filters = {'md': parse_markdown}  # Global filters that modify scalar values

    def deserializer(self):
        """Deserialize file line strings into map object"""
        if self.file_ext == 'md' or self.file_contents:
            self.process_setup()
            if self.file_line_count > 0:
                self.process_line(0, output_data=self.data)
        elif self.file_ext == 'json':
            self.data = self.open_file(self.abspath, rtn_as='json') or {}
        return True

    def process_setup(self):
        lines = self.file_contents or ''
        if self.file_exists and self.file_ext == 'md':
            lines = self.open_file(self.abspath)
        self.file_lines = lines.strip().split("\n")
        self.file_contents = "\n".join(self.file_lines)
        self.file_line_count = len(self.file_lines)

    def process_line(self, cursor, output_data: any = None, is_blob=None, stop_str: str = '') -> tuple:
        """Deserializes string value"""

        def count_tabs(str_value: str):
            """Returns number of tabs for provided string"""
            try:
                return round(len(re.match(r'^\s+', str_value.replace('\n', '')).group()) / 4)
            except Exception as e:
                return 0

        def process_iln_frag(ln_frag, val_type=None):
            """processing inline values for nested objects"""

            def get_pairs(ln_frag):
                """partition key value pairs"""
                # faster than regex! 0.01ms vs 0.45ms
                try:
                    methArgs = ''
                    if "(" in ln_frag:
                        ma = re.search(REG_METH_ARGS, ln_frag)
                        methArgs = ma.group(1)
                        ln_frag = ln_frag.replace(ma.group(), '')

                    iln_delim = [x for x in (
                        (ln_frag.find(STR_DLM), STR_DLM),
                        (ln_frag.find(LST_DLM), LST_DLM),
                        (ln_frag.find(DICT_DELIM), DICT_DELIM),
                    ) if x[0] != -1]
                    return ln_frag.partition(iln_delim[0][1]) + (methArgs,)
                except Exception as e:
                    # print(f"`{ln_frag.strip()}` >>> {e}")
                    return (None, None, ln_frag.strip(), None)

            keystr, delim, valuestr, methargs = get_pairs(ln_frag)

            parsed_key = keystr.strip() if keystr and keystr.strip() != '' else None
            val_type = get_container_type(delim) if val_type is None else val_type
            parsed_val = valuestr.strip()
            force_scalr = delim and delim.endswith('`')
            if parsed_key and parsed_val and not force_scalr:
                has_dotpath = "." in parsed_key
                if has_dotpath or (isinstance(val_type, (list, dict)) or ("," in parsed_val)):  # inline list
                    _c = [] if delim is None else get_container_type(delim)
                    for x in parsed_val.split(','):
                        pk, vtype, pv, pmethArgs = process_iln_frag(x)
                        if vtype != '' and pk: _, pv = self.update_nested(pk, vtype, pv)
                        self.update_nested(None, _c, pv)
                    parsed_val = _c or pv

            return parsed_key, val_type, self.process_value_type(parsed_val), methargs

        def get_container_type(delim):
            if LST_DLM == delim:
                return list()
            elif DICT_DELIM == delim:
                return dict()
            else:
                return str()

        def get_stop(cur, curtabs, is_blob=None, stop_str=None):
            if cur == self.file_line_count: return '__STOPLOOKAHEAD__'
            in_limit = cur + 1 < self.file_line_count
            stop_comm_blok = self.file_lines[cur].strip().endswith(stop_str) if in_limit and stop_str else None
            nxt_curs_is_blok = in_limit and self.file_lines[cur + 1].startswith(BLOCK_PREFIX_STR)
            nxt_curs_tabs = count_tabs(self.file_lines[cur + 1]) if (in_limit and not is_blob) else -1
            return '__STOPLOOKAHEAD__' if stop_comm_blok or nxt_curs_is_blok or (
                    nxt_curs_tabs < curtabs and not is_blob) else None

        stop = False
        stop_iter = False
        while cursor < self.file_line_count:
            self._cursor = cursor
            if stop: break
            ln_frag = self.file_lines[cursor]
            is_multi_ln_comment = ln_frag.strip().startswith('{#')
            is_ln_comment = not is_blob and ln_frag.strip().startswith('#') or not is_blob and ln_frag.strip() == ''
            comment = is_multi_ln_comment or is_ln_comment
            if comment:
                if is_multi_ln_comment or stop_str:
                    cursor, ln_val = self.process_line(cursor + 1, '', stop_str='#}')
            else:
                tabs = count_tabs(ln_frag)
                stop_iter = tabs > 0 and not is_ln_comment or is_blob or stop_str
                try:
                    if is_blob:
                        output_data += ln_frag + "\n"
                    elif not comment and not stop_str:
                        inlimts = cursor + 1 < self.file_line_count
                        is_block = ln_frag.startswith(BLOCK_PREFIX_STR)
                        is_parent = True if is_block else count_tabs(
                            self.file_lines[cursor + 1]) > tabs if inlimts else False
                        parsed_key, val_type, parsed_val, methArgs = process_iln_frag(ln_frag)
                        if methArgs:
                            output_data['@args'] = [arg.replace(' ', '').split(':') for arg in methArgs.split(',')]
                        if is_parent or is_block:
                            parsed_key = parsed_val if not parsed_key else parsed_key
                            parsed_key = "content" if parsed_key == BLOCK_PREFIX_STR else parsed_key.replace(
                                BLOCK_PREFIX_STR, "").strip()
                            cursor, parsed_val = self.process_line(cursor + 1, output_data=val_type,
                                                                   is_blob=isinstance(val_type, str))
                            if isinstance(parsed_val, list) and '-' in parsed_val:  # consolidate list of maps
                                parsed_val = self.post_process_blocklist(parsed_val)
                            if "|" in parsed_key:
                                print('Pipe filters are deprecated', self.abspath)

                        # Store objects with $ prefix
                        if parsed_key and parsed_key.startswith('$'):
                            EmbeddedTypes[parsed_key] = parsed_val
                        else:
                            # Extend objects that inheirit from other files during post processing
                            if parsed_key == '@extends':
                                if not isinstance(parsed_val, dict):
                                    print(f'{self.abspath}')
                                output_data.update(parsed_val)
                                output_data['@extends'] = ln_frag.split(':').pop().strip()
                            else:
                                _, output_data = self.update_nested(parsed_key, output_data, data_merge=parsed_val)
                except Exception as e:
                    # raise Exception(f"{self.file_name}: {str(e)}")
                    raise

            stop = get_stop(cursor, tabs, is_blob, stop_str=stop_str) if stop_iter else None
            if not stop: cursor += 1

        return cursor, output_data

    def process_value_type(self, valuestr: str):
        """Deserialize string value to appropriate object type"""
        if not isinstance(valuestr, str): return valuestr

        def is_num(valstr):
            valstr = valstr.strip().replace(',', '')
            if valstr.isdigit():
                return int(valstr)
            try:
                return float(valstr)
            except ValueError:
                return 'NAN'

        valuestr = valuestr.strip()
        if EmbeddedTypes.get(valuestr):
            return EmbeddedTypes.get(valuestr)

        isnum = is_num(valuestr)
        if isnum != 'NAN':
            return isnum
        if valuestr.strip().lower() == "false":
            return False
        elif valuestr.strip().lower() == "true":
            return True
        elif valuestr.strip().startswith('$'):

            def parse_ref_to_files(filepath, as_dir=0):
                # use proper app context for path reference outside of scope is always the root level
                app_ctx = self.app_ctx  # if not use_root else ("", "", os.path.dirname(self.absdir))
                if as_dir:
                    return allFiles(filepath, force_all=return_all_files, entry_type=None, app_ctx=app_ctx,
                                    **query_params)
                rtn_key = has_attr_path or 'data'
                p = Parsely(filepath, app_ctx)
                d = get_attr(p, rtn_key) or p
                EmbeddedTypes[filepath] = d
                return d

            cvaluestr = valuestr.strip()
            valuestr = valuestr.strip()
            has_file_ref = valuestr.startswith(LOOKUP_FILE_PREFIX)
            has_dir_ref = valuestr.startswith(LOOKUP_DIR_PREFIX)
            query_params = valuestr.split("?").pop() if "?" in valuestr else False
            has_attr_path = valuestr.split("#")[-1] if "#" in valuestr else ''
            valuestr = valuestr.replace(f"{LOOKUP_DIR_PREFIX}/", "") \
                .replace(f"{LOOKUP_FILE_PREFIX}/", "") \
                .replace(f"?{query_params}", "") \
                .replace(f'#{has_attr_path}', '')
            query_params = dict(map(lambda x: x.split("="), query_params.split('&')) if query_params else '')
            use_root = valuestr.startswith('../')
            return_all_files = valuestr.endswith('/*')
            valuestr = valuestr.replace('../', '').replace('/*', '')
            dir_root = os.path.dirname(self.absdir) if self.file_ctx and use_root else self.absdir
            lookup_fpath = os.path.join(dir_root, *valuestr.split("/"))
            if '{{' in lookup_fpath:
                lookup_fpath = self.Filters['jinja'](lookup_fpath, self.data)
            if not os.path.exists(lookup_fpath):
                return self.throw_error({
                    'ISSUE': f'FileNotFound while processing {cvaluestr}',
                    'SOLUTION': f'Make sure the `{lookup_fpath}` file exists. Note that only valid md and json files can be processed.'
                })
            if has_file_ref:
                val = EmbeddedTypes.get(lookup_fpath, parse_ref_to_files(lookup_fpath))
                return val
            elif has_dir_ref:
                return parse_ref_to_files(lookup_fpath, 1)

        return valuestr

    def apply_plugins(self, hook: PyonirHooks):
        """Applies installed active plugins to parsed file"""
        from pyonir import Site
        if Site:
            hook = hook.name.lower()
            for plg in Site.available_plugins:
                if not hasattr(plg, hook): continue
                fn = getattr(plg, hook)
                fn(self, Site)

    def apply_filters(self):
        """Applies filter methods to data attributes"""
        if self.data == None: return
        filters = self.data.get(FILTER_KEY)
        if not filters or self.file_type == 'schemas': return
        for filtr, datakeys in filters.items():
            ifiltr = Parsely.Filters.get(filtr)
            if not ifiltr: continue
            for key in datakeys:
                mod_val = ifiltr(get_attr(self.data, key), self.data)
                self.update_nested(key, self.data, data_update=mod_val)

    @staticmethod
    def serializer(json_map: any, namespace: list = [], inline_mode: bool = False, filter_params=None) -> str:
        """Converts json string into parsely string"""

        if filter_params is None:
            filter_params = {}
        mode = 'INLINE' if inline_mode else 'NESTED'
        lines = []
        multi_line_keys = []
        is_block_str = False

        def pair_map(key, val, tabs):
            is_multiline = isinstance(val, str) and len(val.split("\n")) > 2
            if is_multiline or key in filter_params.get('_blob_keys', []):
                multi_line_keys.append((f"==={key.replace('content', '')}{filter_params.get(key, '')}", val.strip()))
                return
            if mode == 'INLINE':
                ns = ".".join(namespace)
                value = f"{ns}.{key}: {val}" if bool(namespace) else f"{key}: {val.strip()}"
                lines.append(value)
            else:
                # if multiline:
                #     lines.append(f"=== {key}\n{val}")
                if key:
                    lines.append(f"{tabs}{key}: {val}")
                else:
                    lines.append(f"{tabs}{val}")

        if isinstance(json_map, (str, bool, int, float)):
            tabs = '    ' * len(namespace)
            return f"{tabs}{json_map}"

        for k, val in json_map.items():
            tab_count = len(namespace) if namespace is not None else 0
            tabs = '    ' * tab_count
            # block_prefx = BLOCK_PREFIX_STR + ' ' if tab_count == 0 else ''
            # print('\t'*len(namespace), k)
            if isinstance(val, (str, int, bool, float)):
                pair_map(k, val, tabs)

            elif isinstance(val, (dict, list)):
                delim = ':' if isinstance(val, dict) else ':-'
                if len(namespace) > 0:
                    namespace = namespace + [k]
                else:
                    namespace = [k]

                if mode == 'INLINE' and isinstance(val, list):
                    ns = ".".join(namespace)
                    lines.append(f"{ns}{delim}")
                elif mode == 'NESTED':
                    lines.append(f"{tabs}{k}{delim}")

                if isinstance(val, dict):
                    nested_value = Parsely.serializer(json_map=val, namespace=namespace, inline_mode=inline_mode)
                    lines.append(f"{nested_value}")
                else:
                    maxl = len(val) - 1
                    has_scalar = any([isinstance(it, (str, int, float, bool)) for it in val])
                    for i, item in enumerate(val):
                        list_value = Parsely.serializer(json_map=item, namespace=namespace, inline_mode=False)
                        lines.append(f"{list_value}")
                        if i < maxl and not has_scalar:
                            lines.append(f"    -")
                namespace.pop()

        if multi_line_keys:
            [lines.append(f"{mlk}\n{mlv}") for mlk, mlv in multi_line_keys]
        return "\n".join(lines)

    @staticmethod
    def open_file(file_path: str, rtn_as: str = 'string'):
        """Reads target file on file system"""
        if not os.path.exists(file_path): return None
        with open(file_path, 'r', encoding='utf-8') as target_file:
            try:
                if rtn_as == "list":
                    return target_file.readlines()
                elif rtn_as == "json":
                    return json.load(target_file)
                else:
                    return target_file.read()
            except Exception as e:
                return {"error": "pycasso.parsely.Parsely.open_file", "message": str(e)} if rtn_as == "json" else []

    @staticmethod
    def post_process_blocklist(blocklist: list):
        if not isinstance(blocklist, list): return blocklist

        def merge(src, trg):
            ns = []
            for k in src.keys():
                tv = trg.get(k)
                if tv:
                    ns.append(k)
                    trg = trg.get(k)

            Parsely.update_nested(ns, src, trg)
            return src

        _temp_list_obj = {}  # used for blocks that have `-` separated maps
        results = []
        max_count = len(blocklist)
        for i, hashitem in enumerate(blocklist):
            if isinstance(hashitem, dict):
                _temp_list_obj = merge(_temp_list_obj, hashitem)
                if i + 1 == max_count:
                    results.append(dict(_temp_list_obj))
                    break
            else:
                results.append(dict(_temp_list_obj))
                _temp_list_obj.clear()
        blocklist = results
        return blocklist

    @staticmethod
    def update_nested(attr_path: list, data_src: dict, data_merge=None, data_update=None, find=None) -> dict:
        """Finds or Updates target value based on attribute path"""

        def update_value(target, val):
            """updates target object with value parameter"""
            if isinstance(target, list):
                if isinstance(val, list):
                    target += val
                else:
                    target.append(val)
            elif isinstance(target, dict) and isinstance(val, dict):
                target.update(val)
            elif isinstance(target, str) and isinstance(val, str):
                target = val

            return target

        if not attr_path:
            return True, update_value(data_src, data_merge)
        attr_path = attr_path.strip().split('.') if isinstance(attr_path, str) else attr_path
        completed = len(attr_path) == 1
        if isinstance(data_src, list):
            _, data_merge = Parsely.update_nested(attr_path, {}, data_merge)
            return Parsely.update_nested(None, data_src, data_merge)
        elif not completed:
            _data = {}
            for i, k in enumerate(attr_path):
                if find:
                    _data = data_src.get(k) if not _data else _data.get(k)
                else:
                    completed, _data = Parsely.update_nested(attr_path[i + 1:], data_src.get(k, _data), find=find,
                                                             data_merge=data_merge, data_update=data_update)
                    update_value(data_src, {k: _data})
                    if completed: break
        else:
            k = attr_path[-1].strip()
            if find:
                return True, data_src.get(k)
            if data_update: return completed, update_value(data_src, {k: data_update})
            has_mapping_key = isinstance(data_src, (dict,)) and data_src.get(k)
            if not has_mapping_key: data_merge = {k: data_merge}
            update_value(data_src.get(k, data_src), data_merge) if isinstance(data_src, dict) else update_value(
                data_src, data_merge)

        return completed, (data_src if not find else _data)

    @property
    def file_status(self):  # String
        if not self.file_exists: return None
        return 'hidden' if self.file_name.startswith('_') else 'public'

    @property
    def file_created_on(self):  # Datetime
        return datetime.fromtimestamp(os.path.getctime(self.abspath), tz=pytz.UTC) if self.file_exists else None

    @property
    def file_modified_on(self):  # Datetime
        return datetime.fromtimestamp(os.path.getmtime(self.abspath), tz=pytz.UTC) if self.file_exists else None

    @property
    def file_exists(self):
        return os.path.exists(self.abspath) if self.abspath else None

    @property
    def template(self) -> str:
        if not self.file_exists or not isinstance(self.data, dict): return "40x.html"
        context_template = "index.html" if self.is_home else "pages.html"
        return self.data.get('template') or context_template

    @classmethod
    def from_input(cls, input_src: dict, app_ctx: tuple):
        """Creates Parsely object setting the data from input src"""
        if not input_src: return None
        res = cls('', app_ctx)
        res.data = input_src
        return res

    def __init__(self, abspth, app_ctx=None):
        # assert abspth!=None, "Parsely expects an abspath to a resource"
        self.prev_next = None
        ctx_dir, ctx_url, ctx_dirpath, ctx_staticpath = app_ctx
        contents_relpath = abspth.split(ctx_dirpath).pop().lstrip(os.path.sep) if abspth else ''
        contents_rootdir = os.path.dirname(contents_relpath.lstrip(os.path.sep))
        fname, fext = os.path.splitext(os.path.basename(contents_relpath)) if abspth else ('', '')

        self._cursor = None
        self.app_ctx = app_ctx
        self.abspath = abspth
        self.absdir = ctx_dirpath
        self.contents_relpath = contents_relpath
        self.file_ctx = ctx_dir
        self.file_dir = os.path.dirname(contents_relpath)
        self.file_type = contents_rootdir.split(os.path.sep)[0]
        self.file_name = fname
        self.file_ext = fext.split('.').pop()
        self.file_relpath = contents_relpath
        self.file_contents = ''
        self.file_lines = None
        self.file_line_count = None

        # file data processing
        self._blob_keys = []
        self.data = {}
        self.schema = None
        self.deserializer()

        # page attributes
        is_pg = self.file_type != 'schemas'
        surl = re.sub(r'\bpages/\b|\bindex\b', '', contents_relpath.replace(f'.{self.file_ext}', '')) if is_pg else ''
        slug = self.data.get('url', surl.rstrip('/').lstrip('/'))

        self.is_home = not self.file_ctx and self.file_relpath == '/pages/index.md'
        self.slug = '' if self.is_home else f'{ctx_url}/{slug}'.lstrip('/') if is_pg else ''
        self.url = '/' if self.is_home else '/' + self.slug if is_pg else ''

        self.file_ssg_api_dirpath = os.path.join(ctx_staticpath, 'api', self.slug)
        self.file_ssg_html_dirpath = os.path.join(ctx_staticpath, self.slug)
        self.apply_filters()

    def throw_error(self, message: dict):
        msg = {
            'ERROR': f'{self.file_relpath} found an error on line {self._cursor}',
            'LINE': f'{self.file_lines[self._cursor]}', **message}
        return msg

    def set_schema(self, schema_name: str | None = None):
        """Set Schema and schema model for file type or schema_name argument"""
        from pyonir import Site
        schema_path = os.path.join(Site.schemas_dirpath, (schema_name or self.file_type)+'.md')
        schema = Schema(schema_path, self.app_ctx)
        if schema.file_exists:
            self.schema = schema
            self.schema.map_input_to_model(self)

    def set_file_schema(self, schema_name=''):
        """Sets schema model object described by file's type"""
        from . import Site
        if (not self.file_exists and not schema_name): return
        hasSite = hasattr(Site, 'schemas')
        target_schema_name = self.data.get('@schema', schema_name or self.file_type[:-1])
        # Perform lookup schema name against Site schemas
        schema_fpath = os.path.join(self.absdir, 'schemas', target_schema_name + '.md')
        schema_file = get_attr(Site.schemas, target_schema_name) if hasSite else None
        if os.path.exists(schema_fpath):
            schema_file = ParselySchema.from_path(schema_fpath, self.app_ctx)
        plugin_schema_file = None

        self.schema = plugin_schema_file or schema_file

    def refresh_data(self):
        """Parses file and update data values"""
        self.data = {}
        self._blob_keys.clear()
        self.deserializer()

    def output_json(self, data_value: any = None):
        """Outputs a json string"""
        from .utilities import json_serial
        return json.dumps({"data": data_value or self}, default=json_serial)

    def output_html(self, req: PyonirRequest):
        """Renders and html output"""
        from pyonir import Site
        # if not self.is_api and self.file_exists:
        Site.TemplateParser.globals['prevNext'] = self.prev_next
        Site.TemplateParser.globals['page'] = self
        html = Site.TemplateParser.get_template(self.template).render()

        Site.TemplateParser.block_pull_cache.clear()
        return html

    def save(self, file_path=None, contents=None, dir_path=None) -> bool:
        """Saves data into file_path"""
        file_path = file_path or self.abspath
        is_json = file_path.endswith('json')
        if dir_path and not os.path.exists(dir_path):
            os.makedirs(dir_path)
        return create_file(
            file_path,
            contents or self.data,
            is_json=is_json,
            mode='w+')

    def generate_static_file(self, page_request=None, rtn_results=False):
        """Generate target file as html or json. Takes html or json content to save"""
        from pyonir import Site
        # from htmlmin import minify

        count = 0
        html_data = None
        json_data = None

        def render_save():
            # -- Render Content --
            html_data = self.output_html(page_request)
            json_data = self.output_json()
            # -- Save contents --
            self.save(path_to_static_html, html_data, self.file_ssg_html_dirpath)
            self.save(path_to_static_api, json_data, self.file_ssg_api_dirpath)
            return 2

        # -- Get static paths --
        path_to_static_api = os.path.join(self.file_ssg_api_dirpath, "index.json")
        path_to_static_html = os.path.join(self.file_ssg_html_dirpath, "index.html")

        count += render_save()

        if page_request:
            for pgnum in range(1, page_request.paginate):
                path_to_static_html = os.path.join(self.file_ssg_html_dirpath, str(pgnum + 1), "index.html")
                path_to_static_api = os.path.join(self.file_ssg_api_dirpath, str(pgnum + 1), "index.json")
                page_request.query_params['pg'] = pgnum + 1
                count += render_save()

        # -- Return contents without saving --
        if rtn_results:
            return html_data, json_data

        return count


class ParselyPage(Parsely):
    """Page object discovered in the contents/pages directory used to resolve page web request"""

    @property
    def resolver(self):
        return self.data.get('@resolver', None)

    @property
    def entries(self):
        if not self.data or not self.data.get('entries'): return None
        return self.data.get('entries')

    def __init__(self, filepath: str, ctx: str = None):
        from pyonir import PAGINATE_LIMIT
        super().__init__(filepath, app_ctx=ctx)

        self.ctx_endpoint: str = ctx[1] if ctx else None
        self.paginate = None
        self.maxCount = None
        self.pagenum = 1
        self.allow_all = None
        self.show_all = None
        self.lookup = None
        self.limit = PAGINATE_LIMIT
        self.is_resolver = self.data.get('@resolver')

    async def process_response(self, req: PyonirRequest) -> str:
        """process web request into response"""
        from pyonir.server import JSON_RES, TEXT_RES, EVENT_RES, apply_plugin_resolvers
        if not self.file_exists:
            self.data = req.render_error()
        else:
            if self.is_resolver:
                req.type = self.resolver.get('headers', {}).get('accept', req.type)
                return await apply_plugin_resolvers(self, req)
            self.set_paginated_entries(req)
            if req.type == JSON_RES:
                self.set_schema(self.data.get('@schema'))
                return self.output_json(self.schema.model)
        return self.output_html(req) if req.type in (TEXT_RES, '*/*') else self.output_json() \
            if req.type == JSON_RES else self.data

    def set_paginated_entries(self, request: PyonirRequest):
        """Paginated Files contained under a content type directory"""
        limit = int(request.limit)
        pagenum = int(request.query_params.get('pg', 1))
        show_all = None
        entries = Collection(self.data.get('entries', []))
        gallery = Collection(self.gallery.get('files', []), sortBy='file_created_on') if self.gallery else None
        maxCount = entries.count or 0
        limit = int(limit)
        if self.is_api and show_all:
            limit = maxCount
        start = (pagenum * limit) - limit
        end = (limit * pagenum)
        self.limit = limit
        self.pagenum = pagenum
        self.paginate = (maxCount // limit) + (maxCount % limit > 0)
        self.data['entries'] = list(entries.paginate(start=start, end=end, reverse=True))
        if gallery:
            self.data['gallery']['files'] = list(gallery.paginate(start=start, end=end, reverse=True))

    @property
    def is_api(self):
        return self.file_dir == 'api'

    @property
    def name(self):
        return self.data.get('name', self.data.get('title', self.file_name))

    @property
    def title(self):
        return self.data.get('title', self.name)

    @property
    def date(self):
        """Returns manual published date but defaults to file_created_on date"""
        import datetime
        file_date = self.data.get('date', self.file_created_on) if isinstance(self.data, dict) else self.file_created_on
        event_date = Collection.get_attr(self.data, 'event.begins.start_datetime')
        if isinstance(file_date, str):
            file_date = deserialize_datestr(file_date)
        return event_date if isinstance(event_date, datetime.datetime) else file_date

    @property
    def created_on(self):
        return self.file_created_on

    @property
    def modified_on(self):
        return deserialize_datestr(self.data.get('@last_saved', self.file_modified_on))

    @property
    def excerpt(self):
        """Summary of content in 280 characters"""
        if not self.data or not isinstance(self.data, dict): return None
        default_excerpt = remove_html_tags(self.content or '')[0:280] + "..."
        return self.data.get('excerpt', default_excerpt)

    @property
    def content(self):
        return self.data.get('content',
                             ' ')  # self.markdown_parser(self.data.get('content')) if self.data.get('content') else ''

    @property
    def canonical(self):
        from pyonir import Site
        return f"{Site.configs.site.domain}{self.url}"

    @property
    def tags(self):
        if not self.file_exists: return None
        catlist = os.path.dirname(self.file_relpath).lstrip('/').split('/')
        return self.data.get('tags', '').split(',') if self.data.get('tags') else catlist

    @property
    def category(self) -> str:
        if not self.file_exists: return None
        catlist = os.path.dirname(self.file_relpath).lstrip('/').split('/')
        return self.data.get('category', catlist[-1])

    @property
    def order(self):
        return self.data.get('order', 99)

    @property
    def author(self):
        return self.data.get('author')

    @property
    def status(self):
        return self.data.get('status', self.file_status)

    @property
    def menu(self):
        if not self.data: return None
        return self.data.get('menu')

    @property
    def menu_parent(self):
        if not self.data: return None
        return self.data.get('menu_parent', None)

    @property
    def model(self):
        """overrides page schema model name"""
        return self.data.get('page_schema', None) if self.data else None

    @property
    def redirect_url(self):
        return self.data.get('redirect_url') if self.data else None

    @property
    def js(self):
        """Javascript content block"""
        if not self.file_exists or not self.data.get('js'): return None
        js_str = self.data.get('js') if isinstance(self.data.get('js'), str) else None
        jsdata = {"script": js_str}
        if isinstance(self.data.get('js'), dict):
            jsdata.update(**self.data.get('js'))
        return jsdata

    @property
    def css(self):
        """CSS styles to include on html template"""
        if not self.file_exists or not self.data.get('css'): return None
        css_str = self.data.get('css') if isinstance(self.data.get('css'), str) else None
        cssdata = {"style": css_str}
        if isinstance(self.data.get('css'), dict):
            cssdata.update(**self.data.get('css'))
        return tupleconverter('css', cssdata)

    @property
    def gallery(self):
        """Gallery plugin object"""
        return self.data.get('gallery')

    @property
    def form(self):
        """Form plugin object"""
        return self.data.get('form')

    @property
    def ctx_redirect(self):
        return self.data.get('redirect') if isinstance(self.data, dict) else None

    def prev_next(self):
        """Returns sorted dictionary for prev and next page based on date specified in file"""
        if self.file_type != 'pages' or self.is_home: return None

        def pgen():
            for item in Collection(allFiles(os.path.dirname(self.abspath))).collection:
                yield item

        # result = {"next": False, "prev": False}
        prv = None
        nxt = None
        home = {'title': self.category, 'url': os.path.dirname(self.url)}
        gen_data = pgen()
        for file in gen_data:
            if file.file_status == 'hidden' or not file.file_exists: continue
            if self.file_name == file.file_name:
                try:
                    nxt = next(gen_data)
                    break
                except StopIteration:
                    if nxt:
                        nxt.set_file_schema()
                        nxt = nxt.file_schema.map_to_schema(nxt) if nxt.file_schema else nxt
                    if prv:
                        prv.set_file_schema()
                        prv = prv.file_schema.map_to_schema(prv) if prv.file_schema else prv
                    pass
            else:
                prv = file

        return {"next": nxt or home, "prev": prv or home, "home": home}


class ParselyMedia(Parsely):
    """Represents an image or media document file"""

    @property
    def thumbnails(self):
        """Returns a map of thumbnail sizes available for media file"""
        return self.data.get('thumbnails')

    @property
    def id(self):
        from pyonir import UPLOADS_THUMBNAIL_DIRNAME, UPLOADS_DIRNAME
        if not self.data: return None
        if self.is_thumb:
            return self.data.get('full_url')
        group = self.data.get('group')
        if group != UPLOADS_DIRNAME: group = f'{UPLOADS_DIRNAME}/{group}'
        name = self.name.split(' ', 1)[0].lower()
        return f"{group}/{name}.{self.file_ext}"

    def __init__(self, file_path: str = None, app_ctx=None):
        super().__init__(file_path, app_ctx)
        from pyonir import UPLOADS_ROUTE

        self.sizes = []
        self.url = f"{UPLOADS_ROUTE}{self.file_relpath.replace(self.file_type, '')}"
        self.slug = self.url.lstrip('/')
        self.is_img = None
        self.raw_img = None
        self.name = ''
        self.is_thumb = None
        self.set_data()

    def set_data(self):
        from PIL import Image
        from pyonir import UPLOADS_DIRNAME, UPLOADS_THUMBNAIL_DIRNAME
        if not self.file_exists: return
        image_name, *image_captions = self.file_name.replace('.' + self.file_ext, '') \
            .split(IMG_FILENAME_DELIM)

        is_thumb = UPLOADS_THUMBNAIL_DIRNAME in self.file_dir
        formatted_name = re.sub(r'[^a-zA-Z0-9]+', ' ', image_name).title()
        formated_caption = "".join(image_captions or self.name).title()
        full_img = self.slug.replace(UPLOADS_THUMBNAIL_DIRNAME + '/', '').split('--')[0] + f".{self.file_ext}" \
            if is_thumb else self.url
        self.raw_img = Image.open(self.abspath)
        self.name = formatted_name
        self.is_thumb = is_thumb
        self.data = {
            "@schema": "media",
            "file_name": image_name if not is_thumb else image_name.split('--')[0],
            "name": self.name,
            "url": self.url,
            "full_url": full_img,
            "slug": self.slug,
            "date": self.file_created_on,
            "type": self.file_ext,
            "group": self.file_dir.split(f'{UPLOADS_DIRNAME}{os.path.sep}', 1).pop(),
            "captions": formated_caption.title(),
            "size": os.path.getsize(self.abspath),
            "width": get_attr(self.raw_img, "width", None),
            "height": get_attr(self.raw_img, "height", None)
        }
        self.data.update({"thumbnails": self.get_all_thumbnails()})

    def get_all_thumbnails(self) -> dict | None:
        """Collects thumbnails for the image"""
        if self.is_thumb: return None
        from pyonir import UPLOADS_DIRNAME, UPLOADS_THUMBNAIL_DIRNAME
        group_dir = self.data.get('group')
        if group_dir != UPLOADS_DIRNAME: group_dir = f'{UPLOADS_DIRNAME}/{group_dir}'
        thumbs_dir = os.path.join(self.absdir, group_dir, UPLOADS_THUMBNAIL_DIRNAME)
        files = allFiles(thumbs_dir, app_ctx=self.app_ctx)
        target_name = self.data.get('file_name')
        thumbs = {}
        # filter files based on name
        for file in files:
            if file.data.get('file_name') != target_name: continue
            w = file.data.get('width')
            h = file.data.get('height')
            thumbs[f'{w}x{h}'] = file
            pass
        return thumbs

    @staticmethod
    def createImagefolders(folderpath: str):
        thumbspath = os.path.join(folderpath, 'thumbnails')
        if not os.path.exists(folderpath):
            os.makedirs(folderpath)
        if not os.path.exists(thumbspath):
            os.makedirs(thumbspath)

    @staticmethod
    async def save_upload(upload_doc, img_folder_abspath) -> 'ParselyMedia':
        """Saves base64 file contents into file system"""
        filename, filedata, rootpath = upload_doc
        file_name, file_ext = os.path.splitext(filename)
        ParselyMedia.createImagefolders(img_folder_abspath)
        abspath = os.path.join(img_folder_abspath, file_name + file_ext)
        file_contents = await filedata.read()
        with open(abspath, 'wb') as f:
            f.write(file_contents)
        parselyMedia = ParselyMedia.new(abspath)
        return parselyMedia

    @classmethod
    def new(cls, file_path):
        from pyonir import Site
        pm = cls(file_path, Site.files_ctx)
        return pm

    def resize(self, sizes=None):
        '''
        Resize each image and save to the upload path in corresponding image size and paths
        This happens after full size images are saved to the filesystem
        '''
        from PIL import Image
        from pyonir import THUMBNAIL_DEFAULT
        if sizes is None:
            sizes = [THUMBNAIL_DEFAULT]
        try:
            for dimensions in sizes:
                width, height = dimensions
                self.sizes.append(dimensions)
                img = self.raw_img.resize((width, height), Image.Resampling.BICUBIC)
                file_name = f'{self.file_name}--{width}x{height}'
                folderpath = os.path.dirname(self.abspath)
                self.createImagefolders(folderpath)
                filepath = os.path.join(folderpath, 'thumbnails', file_name + '.' + self.file_ext)
                if not os.path.exists(filepath): img.save(filepath)
        except Exception as e:
            raise

    def generate_thumb(self, width, height) -> str:
        """Generates an image accordion to width and height parameters and returns url to the new resized image"""
        self.sizes.append((width, height))
        if not self.thumbnails.get(f'{width}x{height}'): self.resize([(width, height)])
        return self.thumbnails.get(f'{width}x{height}')


class Schema(Parsely):
    PRIVATE_PREFIX = '_'

    @property
    def provider_model(self) -> dict | None:
        """Public model field names"""
        return self.data.get('@provider_model', {})

    @property
    def primary_key(self):
        """Primary key column name"""
        return self.data.get('@pk', 'id')

    def __init__(self, schema_path: str, app_ctx: list):
        super().__init__(schema_path, app_ctx)
        self.fields: set[str] = set(Schema.sanitize(k) for k in self.data.keys() if not k.startswith('@'))\
            if self.file_exists else None
        self.private_fields = set(Schema.sanitize(f) for f in self.data.keys() if f.startswith('_'))
        self.model = None
        pass

    @staticmethod
    def sanitize(key: str):
        return key[1:] if key.startswith(Schema.PRIVATE_PREFIX) else key

    def map_input_to_model(self, subject: dict | Parsely):
        """Returns a schema model from subject data"""
        def get_value(key):
            """get key value on subject"""
            v = get_attr(subject, key)
            v = get_attr(subject.data, key) if v is None and isinstance(subject, Parsely) else v
            if isinstance(self.data.get(key), (list, set, Generator)):
                itr_schema_type = self.data.get(key)[0]
                entries = []
                for entry in v:
                    if hasattr(entry,'set_schema'): entry.set_schema(itr_schema_type)
                    target = entry.schema.model if hasattr(entry,'set_schema') else entry
                    entries.append(target)
                v = entries
            return "***" if key in self.private_fields and v is not None else v

        model = type(self.file_name, (object,), {"__schema__": self.file_name})()
        for field in self.fields:
            v = get_value(field)
            setattr(model, self.provider_model.get(field, field), v)
            pass
        self.model = model
        return model

    def validate_model(self):
        """Validates model against schema conditions"""
        pass


class SchemaValidator:
    """Collection of functions to validate schema values"""

    @property
    def validation_errors(self):
        return list(self.errors_bag.values()) if bool(self.errors_bag) else None

    @property
    def is_valid(self):
        return True if len(self.errors_bag) == 0 or self.schema is None else False

    def __init__(self, schema):
        self.schema = schema
        self.__schema_value = None
        self.__schema_key = None
        self.errors_bag = dict()

    def send_error(self, msg=None):
        self.errors_bag[self.__schema_key] = f"{msg or 'error found'}"

    def validate_input(self, input_data):
        """Validates input data against available schema definitions"""
        self.errors_bag.clear()
        for field in self.schema.required_fields:
            schema_methods = self.schema.file.data.get(f"*{field}")
            self.__schema_key = field
            self.__schema_value = get_attr(input_data, self.__schema_key)
            for meth in schema_methods.split('|'):
                if getattr(self, meth.strip())() is True: continue
                break
        return self

    def update(self, schema):
        self.schema = schema

    def bool(self):
        return isinstance(self.__schema_value, bool)

    def is_int(self):
        return isinstance(self.__schema_value, int) or isinstance(self.__schema_value, float)

    def string(self):
        if isinstance(self.__schema_value, str):
            return True
        else:
            self.send_error(f"{self.__schema_key} is not a string! Got an {type(self.__schema_value).__name__}")

    def is_date(self):
        import datetime
        date_obj = deserialize_datestr(self.__schema_value, auto_correct=False)
        if not isinstance(date_obj, datetime.datetime):
            self.send_error(f"{self.__schema_value} is not a valid date value. {date_obj}")
        else:
            return True

    def is_list(self):
        return isinstance(self.__schema_value, list)

    def as_address(self):
        try:
            street, city, state_zip = self.__schema_value.split(',')
            state, zip = state_zip.strip().split(' ')
            self.input_data[self.__schema_key] = {
                "street": street.strip(),
                "city": city.strip(),
                "state": state.strip(),
                "zip": zip.strip()
            }
            return True
        except Exception as e:
            self.send_error(str(e))
            return False

    def check_role(self):
        return True

    def check_email(self, email=None):
        try:
            from django.core.validators import validate_email, ValidationError
            validate_email(email or self.__schema_value)
            return True
        except ValidationError:
            self.send_error(f"{self.__schema_key.capitalize()} is not a valid.")
            # return 'Email is not valid.'
            return False

    def min(self, min_num):
        has_minimum = len(self.__schema_value) >= int(min_num)
        if not has_minimum:
            self.send_error(f"{self.__schema_key} should have a minimum of {min_num} characters.")
        return has_minimum

    def check_pwd(self):
        # from pycasso.utils import Collection
        # pswd = password or Collection.get_attr(self.schema.input_data, 'password', True)
        error = 'Password requires alpha numeric values with 6 characters.'
        if not self.__schema_value:
            error = 'Must provide a password.'
        elif not self.__schema_value.isdigit() and len(self.__schema_value) > 5:
            return True
        self.send_error(error)
        return False

    def sanitize(self):
        illegal_chars = "<,>,!,+,-,(,),~,`,\,/,:,;,&,^,%,$,#".split(',')
        for c in self.__schema_value:
            if c in illegal_chars:
                self.send_error(f"{self.__schema_key} contains an invalid character.")
                return False
        return True  # if self.__schema_value not in illegal_chars else False



def valueFrom(rowObj, path=""):
    targetObj = None
    for key in path.split('.'):
        try:
            if targetObj:
                targetObj = targetObj[key]
            else:
                targetObj = rowObj[key]
            pass
        except Exception as error:
            raise error
    return targetObj


# -*- coding: utf-8 -*-
import os, sys
from pyonir import utilities
from pyonir.parser import Parsely
from pyonir.libs.plugins.forms import Forms
from pyonir.libs.plugins.navigation import Navigation
from pyonir.types import PyonirHooks

SYS_USER = os.path.basename(os.environ.get('HOME'))
# Pycasso settings
PYONIR_DIRPATH = os.path.abspath(os.path.dirname(__file__))
PYONIR_LIBS_DIRPATH = os.path.join(PYONIR_DIRPATH, "libs")
PYONIR_PLUGINS_DIRPATH = os.path.join(PYONIR_LIBS_DIRPATH, 'plugins')
PYONIR_SETUPS_DIRPATH = os.path.join(PYONIR_LIBS_DIRPATH, 'app_setup')
PYONIR_JINJA_DIRPATH = os.path.join(PYONIR_LIBS_DIRPATH, 'jinja')
PYONIR_JINJA_TEMPLATES_DIRPATH = os.path.join(PYONIR_JINJA_DIRPATH, "templates")
PYONIR_JINJA_EXTS_DIRPATH = os.path.join(PYONIR_JINJA_DIRPATH, "extensions")
PYONIR_JINJA_FILTERS_DIRPATH = os.path.join(PYONIR_JINJA_DIRPATH, "filters")
PYONIR_MESSAGES_FILE = os.path.join(PYONIR_LIBS_DIRPATH, "system-messages.md")
PYONIR_SSL_KEY = os.path.join(PYONIR_SETUPS_DIRPATH, "content/certs/server.key")
PYONIR_SSL_CRT = os.path.join(PYONIR_SETUPS_DIRPATH, "content/certs/server.crt")
PYONIR_STATIC_ROUTE = "/pyonir_assets"
PYONIR_STATIC_DIRPATH = os.path.join(PYONIR_LIBS_DIRPATH, 'ui-kits', 'static')

# Environments
DEV_ENV = 'LOCAL'
STAGE_ENV = 'STAGING'
PROD_ENV = 'PROD'

###########################
#     System Config       #
###########################
EXTENSIONS = {"file": ".md", "settings": ".json"}
THUMBNAIL_DEFAULT = (230, 350)
PROTECTED_FILES = ('.', '_', '<', '>', '(', ')', '$', '!', '._',)
IGNORE_FILES = ('.vscode', '.vs', '.DS_Store', '__pycache__', '.git')

PAGINATE_LIMIT = 6
RECENT_POSTS = 2
DATE_FORMAT = "%Y-%m-%d %I:%M:%S %p"
TIMEZONE = "US/Eastern"
ALLOWED_UPLOAD_EXTENSIONS = ('jpg', 'JPG', 'PNG', 'png', 'txt', 'md', 'jpeg', 'pdf', 'svg', 'gif')

# Base application directories
APPS_DIRNAME = "apps"  # dirname for any child apps
SITES_DIRNAME = "sites"  # dirname for any sites
BACKEND_DIRNAME = "backend"  # dirname for all backend markdown files
FRONTEND_DIRNAME = "frontend"  # dirname for all html templates and themes
CONTENTS_DIRNAME = "contents"  # dirname for site markdown data

# Backend sub directories
CONTROLLERS_DIRNAME = "controllers"
PLUGINS_DIRNAME = "plugins"
FILTERS_DIRNAME = "filters"
BACKEND_SUBDIRS = (
    f"{BACKEND_DIRNAME}/{CONTROLLERS_DIRNAME}",
    f"{BACKEND_DIRNAME}/{PLUGINS_DIRNAME}"
)

# Content sub directories
API_DIRNAME = "api"
DATASTORE_DIRNAME = "datastore"
PAGES_DIRNAME = "pages"
UPLOADS_DIRNAME = "uploads"
SCHEMAS_DIRNAME = "schemas"
CERTS_DIRNAME = "certs"
USERS_DIRNAME = "users"
CONFIGS_DIRNAME = "configs"
FORMS_DIRNAME = "forms"
LOGS_DIRNAME = "logs"
CONTENTS_DIRNAMES = (
    API_DIRNAME, PAGES_DIRNAME, CONFIGS_DIRNAME, SCHEMAS_DIRNAME, DATASTORE_DIRNAME, FORMS_DIRNAME, UPLOADS_DIRNAME)
CONTENT_SUBDIRS = (
    f"{CONTENTS_DIRNAME}/{API_DIRNAME}", CONTENTS_DIRNAME + "/" + CERTS_DIRNAME,
    CONTENTS_DIRNAME + "/" + CONFIGS_DIRNAME, CONTENTS_DIRNAME + "/" + UPLOADS_DIRNAME,
    CONTENTS_DIRNAME + "/" + SCHEMAS_DIRNAME, CONTENTS_DIRNAME + "/" + FORMS_DIRNAME,
    CONTENTS_DIRNAME + "/" + DATASTORE_DIRNAME, CONTENTS_DIRNAME + "/" + PAGES_DIRNAME
)

APPLICATION_DIRS = (BACKEND_DIRNAME, FRONTEND_DIRNAME, CONTENTS_DIRNAME) + BACKEND_SUBDIRS + CONTENT_SUBDIRS

# Frontend
THEMES_DIRNAME = "themes"
THEME_STATIC_DIRNAME = "static"
THEME_TEMPLATES_DIRNAME = "layouts"
THEME_NAME = "pencil"
ASSETS_DIRNAME = "public"
SSG_DIRNAME = "static_site"
FRONTEND_SUBDIRS = (THEMES_DIRNAME, SSG_DIRNAME)

# Routes
API_ROUTE = f"/{API_DIRNAME}"  # Api base path for accessing pages as JSON
UPLOADS_ROUTE = f"/{UPLOADS_DIRNAME}"  # Upload base path to access resources within upload directory
ASSETS_ROUTE = f"/{ASSETS_DIRNAME}"  # serves static assets from configured theme
UPLOADS_THUMBNAIL_DIRNAME = "thumbnails"  # thumbnail sub directory name

VERSION = os.path.basename(os.path.dirname(os.path.abspath(os.path.dirname(__file__))))
HOST = 'localhost'  # '0.0.0.0'
PORT = 5000
__version__ = '1.0.0'
Site: 'IApp' = None


def init(site_path: str):
    """Initializes existing Pyonir application"""
    global Site
    # Set Global Site instance
    Site = IApp(site_path)
    return Site


class IApp:
    """Main Pyonir application"""

    SECRET_SAUCE = 'F12Zr47j\3yX R~X@H!jmM]Lwf/,?KT'  # used for salting tokens
    SESSION_KEY = 'pyonir_jwt'  # used for salting tokens
    VERSION = 'beta'  # running application version
    SYS_VERSION = VERSION
    SYS_USER = SYS_USER
    SYS_STATIC_DIRPATH = PYONIR_STATIC_DIRPATH
    endpoint = ''  # Home application

    @staticmethod
    def register_parsely_extension(plugins: list = None):
        """Registers new file parsing plugins"""
        if not plugins: return
        for plg in plugins:
            # Install parsely extension
            if hasattr(plg, 'parsely_extension'):
                Parsely.Extensions[plg.parsely_extension] = plg
        pass

    @property
    def env_home(self):
        """Returns value found in bash_profile defined for APPENV"""
        return os.environ.get('APPHOME')

    @property
    def env(self):
        """Returns value found in bash_profile defined for APPENV"""
        return os.getenv('APPENV')

    @property
    def is_dev(self):
        """Status returns if application is running in local dev mode"""
        return self.env == DEV_ENV

    @property
    def site_logs_dirpath(self):
        """Nginx conf file install location"""
        return os.path.join(self.app_dirpath, "logs")

    @property
    def app_socket_filename(self):
        """WSGI or ASGI file name"""
        return f"{self.name}.sock"

    @property
    def app_socket_filepath(self):
        """WSGI or ASGI file install location"""
        return os.path.join(self.app_dirpath, self.app_socket_filename)

    @property
    def app_nginx_conf_filepath(self):
        """NGINX configuration file"""
        return os.path.join(self.app_dirpath, self.name + '.conf')

    @property
    def app_sitemap_filepath(self):
        """WSGI or ASGI file install location"""
        return os.path.join(self.theme_static_dirpath, f"sitemap.xml")

    @property
    def domain(self):
        is_prod = (self.env is not DEV_ENV) and getattr(self.configs.app, 'domain', None)
        return self.configs.app.domain if is_prod else HOST

    @property
    def host(self):
        return self.domain

    @property
    def protocol(self):
        return 'https' if self.is_secure else 'http'

    @property
    def origin(self):
        addr = f"{self.protocol}://{self.domain}"
        return f"{addr}:{self.port}" if self.env is DEV_ENV else addr

    @property
    def port(self):
        """Port address for application"""
        return self.configs.port if hasattr(self.configs, 'port') else PORT

    @property
    def is_secure(self):
        return getattr(self.configs.app, 'use_ssl', None)

    @property
    def account_dirpath(self):
        """Site Account dirname all applications are install"""
        return os.path.dirname(self.app_dirpath)

    @property
    def account_dirname(self):
        """Site Account dirpath all applications are install"""
        return os.path.basename(os.path.dirname(self.app_dirpath))

    @property
    def themes(self):
        """Available themes installed in the frontend/themes directory"""
        return [Parsely(os.path.join(self.themes_dirpath, dirname, 'README.md')).data for dirname in
                os.listdir(self.themes_dirpath)
                if os.path.exists(os.path.join(self.themes_dirpath, dirname, 'README.md'))]

    @property
    def theme(self):
        """Active theme name displayed for website"""
        if hasattr(self, 'configs') and not hasattr(self.configs.app, 'theme_name'): return
        return self.configs.app.theme_name

    @property
    def theme_templates_dirpath(self):
        """Active Theme templates dirpath to access html files used to render a page during web requests"""
        if not self.theme: return
        return os.path.join(self.frontend_dirpath, THEMES_DIRNAME, self.theme, THEME_TEMPLATES_DIRNAME)

    @property
    def theme_layouts_and_files(self):
        """Map of all layout files and macros available inside a themes directory"""
        return os.listdir(self.theme_templates_dirpath)

    @property
    def theme_static_dirpath(self):
        """Active Theme static assets dirpath to access js,css and other web resources"""
        if not self.theme: return
        return os.path.join(self.frontend_dirpath, THEMES_DIRNAME, self.theme, THEME_STATIC_DIRNAME)

    @property
    def theme_assets_dirpath(self):
        """Active Theme static assets dirpath to access js,css and other web resources"""
        if not self.theme: return
        return os.path.join(self.theme_static_dirpath, ASSETS_DIRNAME)

    @property
    def ssg_dirpath(self):
        """Static site output directory path"""
        return os.path.join(self.app_dirpath, SSG_DIRNAME)

    @property
    def ssl_public_cert(self):
        """Static site output directory path"""
        pub_cert = utilities.Collection.get_attr(self.configs.app, 'public_cert') or 'server.crt'
        return os.path.join(self.contents_dirpath, CERTS_DIRNAME, pub_cert)

    @property
    def ssl_private_cert(self):
        """Static site output directory path"""
        priv_cert = utilities.Collection.get_attr(self.configs.app, 'private_cert') or 'server.key'
        return os.path.join(self.contents_dirpath, CERTS_DIRNAME, priv_cert)

    @property
    def request_paths(self):
        """list of base file paths to resolve web requests. includes child apps"""
        _ppth = [self.pages_dirpath, self.api_dirpath]
        _prunes = [os.path.basename(p) for p in _ppth]
        return [(self.name, '/', _ppth, _prunes)]

    def __init__(self, app_rootpth: str = None) -> None:
        self.server = None
        Parsely.Filters['jinja'] = self.parse_jinja
        self.name = os.path.basename(app_rootpth)
        self.app_dirpath = app_rootpth
        self.app_entrypoint = os.path.join(self.app_dirpath, 'main.py')
        self.logs_dirpath = os.path.join(app_rootpth, LOGS_DIRNAME)
        self.backend_dirpath = os.path.join(app_rootpth, BACKEND_DIRNAME)
        self.plugins_dirpath = os.path.join(self.backend_dirpath, PLUGINS_DIRNAME)
        self.filters_dirpath = os.path.join(self.backend_dirpath, FILTERS_DIRNAME)
        self.contents_dirpath = os.path.join(self.backend_dirpath, CONTENTS_DIRNAME)
        self.api_dirpath = os.path.join(self.contents_dirpath, API_DIRNAME)
        self.users_dirpath = os.path.join(self.contents_dirpath, USERS_DIRNAME)
        self.pages_dirpath = os.path.join(self.contents_dirpath, PAGES_DIRNAME)
        self.schemas_dirpath = os.path.join(self.contents_dirpath, SCHEMAS_DIRNAME)
        self.uploads_dirpath = os.path.join(self.contents_dirpath, UPLOADS_DIRNAME)
        self.datastore_dirpath = os.path.join(self.contents_dirpath, DATASTORE_DIRNAME)
        self.frontend_dirpath = os.path.join(app_rootpth, FRONTEND_DIRNAME)
        self.themes_dirpath = os.path.join(self.frontend_dirpath, THEMES_DIRNAME)
        self.files_ctx = ('', '', self.contents_dirpath, self.ssg_dirpath)
        self.messages = {}
        self.TemplateParser = None
        self.configs = self.process_contents(os.path.join(self.contents_dirpath, CONFIGS_DIRNAME), self.files_ctx)
        self.jinja_template_dirpaths = (self.theme_templates_dirpath, PYONIR_JINJA_TEMPLATES_DIRPATH,)
        self.setup_jinja()
        # self.schemas = self.process_contents(os.path.join(self.contents_dirpath, SCHEMAS_DIRNAME), self.files_ctx)

        # Setups
        self.setup_system_msgs()
        self.available_plugins = {Forms(self), Navigation(self)}

    async def run_plugins(self, hook: PyonirHooks, data_value=None):
        if not hook or not self.available_plugins: return
        hook = hook.name.lower()
        for plg in self.available_plugins:
            if not hasattr(plg, hook): continue
            hook_method = getattr(plg, hook)
            await hook_method(data_value, self)

    def parse_jinja(self, string, context={}) -> str:
        try:
            return self.TemplateParser.from_string(string).render(configs=self.configs, **context)
        except Exception as e:
            raise

    def setup_system_msgs(self):
        """App and Pycasso system messages for logs and user display"""
        sys_msgs = Parsely(PYONIR_MESSAGES_FILE, self.files_ctx)
        app_msgs = Parsely(os.path.join(self.datastore_dirpath, 'app-messages.md'), self.files_ctx)
        self.messages.update(sys_msgs.data)
        self.messages.update(app_msgs.data)
        pass

    def setup_jinja(self):
        """Jinja parser used to generate html from jinja templates"""
        from jinja2 import Environment, FileSystemLoader
        from webassets import Environment as AssetsEnvironment
        from webassets.ext.jinja2 import AssetsExtension

        def url_for(path):
            rmaps = self.server.url_map
            return rmaps.get(path, {}).get('path', ASSETS_ROUTE)

        sysfilters = self.module_loader(PYONIR_JINJA_FILTERS_DIRPATH, default_data={})
        appfilters = self.module_loader(self.filters_dirpath, default_data={})
        self.JINJA_INSTALLED_FILTERS = {**sysfilters, **appfilters}
        self.JINJA_INSTALLED_EXTENSIONS = self.module_loader(PYONIR_JINJA_EXTS_DIRPATH, default_data=[],
                                                             rtn_attr='import_str')

        app_extensions = [AssetsExtension]
        for ext_module_str in self.JINJA_INSTALLED_EXTENSIONS:
            app_extensions.append(ext_module_str)

        self.TemplateParser = Environment(
            # '''path to find templates'''
            loader=FileSystemLoader(self.jinja_template_dirpaths),
            # '''include custom extensions'''
            extensions=app_extensions,
            # '''disable reloading of template changes'''
            auto_reload=True,
        )

        #  ''' Custom filters '''
        self.TemplateParser.filters.update(**self.JINJA_INSTALLED_FILTERS)

        # load assests tag
        self.TemplateParser.assets_environment = AssetsEnvironment(self.theme_static_dirpath, ASSETS_ROUTE)
        # Add paths containing static assets
        self.TemplateParser.assets_environment.load_path.append(self.theme_static_dirpath)
        self.TemplateParser.url_expire = False
        self.TemplateParser.globals['request'] = {}
        self.TemplateParser.globals['configs'] = self.configs.app
        self.TemplateParser.globals['url_for'] = url_for

    def generate_static_website(self):
        """Generates Static website into the specified static_site_dirpath"""
        import time
        from pyonir.server import generate_nginx_conf

        self.SSG_INPROGRESS = True
        count = 0
        print(f"{utilities.pcolors.OKBLUE}1. Coping Assets")
        try:

            static_site_dirpath = os.path.join(self.app_dirpath, SSG_DIRNAME)
            generate_nginx_conf(self)
            print(f"{utilities.pcolors.OKCYAN}3. Generating Static Pages")

            self.TemplateParser.globals['is_ssg'] = True
            start_time = time.perf_counter()

            all_pages = utilities.allFiles(self.pages_dirpath, app_ctx=self.files_ctx)
            xmls = []
            for page in all_pages:
                # pg_req = Request(self, page)
                self.TemplateParser.globals['request'] = page  # pg_req
                count += page.generate_static_file()
                t = f"<url><loc>{self.protocol}://{self.domain}{page.url}</loc><priority>1.0</priority></url>\n"
                xmls.append(t)
                self.TemplateParser.block_pull_cache.clear()
            del self.TemplateParser.globals['is_ssg']
            smap = f'<?xml version="1.0" encoding="UTF-8"?><urlset xmlns="http://www.sitemaps.org/schemas/sitemap/0.9"><url><loc>{self.domain}</loc><priority>1.0</priority></url> {"".join(xmls)} </urlset>'
            utilities.create_file(self.app_sitemap_filepath, smap, 0)
            utilities.copy_assets(PYONIR_STATIC_DIRPATH,
                                  os.path.join(static_site_dirpath, PYONIR_STATIC_ROUTE.lstrip('/')))
            utilities.copy_assets(self.theme_static_dirpath, os.path.join(static_site_dirpath, ASSETS_DIRNAME))
            end_time = time.perf_counter() - start_time
            ms = end_time * 1000
            count += 3
            msg = f"SSG generated {count} html/json files in {round(end_time, 2)} secs :  {round(ms, 2)} ms"
            print(f'\033[95m {msg}')
        except Exception as e:
            msg = f"SSG encountered an error: {str(e)}"
            raise

        self.SSG_INPROGRESS = False
        response = {"status": "COMPLETE", "msg": msg, "files": count}
        print(response)
        print(utilities.pcolors.RESET)
        return response

    def run(self, endpoints: 'Endpoints'):
        """Runs the Uvicorn webserver"""
        from .server import (setup_starlette_server, start_uvicorn_server,
                             pyonir_file_upload,pyonir_file_delete, pyonir_sse_handler)
        # Initialize Server instance
        self.server = setup_starlette_server(self)
        self.server.resolvers[pyonir_file_upload.__name__] = pyonir_file_upload
        self.server.resolvers[pyonir_file_delete.__name__] = pyonir_file_delete
        self.server.resolvers[pyonir_sse_handler.__name__] = pyonir_sse_handler

        # Run uvicorn server
        start_uvicorn_server(self, endpoints)

    @staticmethod
    def module_loader(path_to_modules, default_data=None, target_module_name=None, rtn_attr="", as_pkg=False) -> any:
        """
        This will preload python files from the provided modules path.
        Pycasso system and applications .py files will be accessible at runtime.
        """
        if default_data is None:
            default_data = []
        if not os.path.exists(path_to_modules): return default_data
        try:
            for module_dirname in os.listdir(path_to_modules):
                if module_dirname in IGNORE_FILES or \
                        module_dirname.startswith(PROTECTED_FILES): continue
                # convert path into module object
                module_dirname = os.path.join(module_dirname,
                                              target_module_name) if target_module_name else module_dirname
                if not os.path.exists(os.path.join(path_to_modules, module_dirname)): continue
                if not module_dirname.endswith('.py') and not as_pkg: continue
                pkg = utilities.PyonirMod(os.path.join(path_to_modules, module_dirname))
                if not pkg.class_module: continue
                name = f"{pkg.import_str}" if target_module_name else pkg.callable_name
                mod = getattr(pkg, rtn_attr, None)
                if isinstance(default_data, dict):
                    default_data.update(
                        {name: mod or pkg.class_module}
                    )
                elif isinstance(default_data, list):
                    default_data.append(mod or pkg)
        except Exception as e:
            raise
        return default_data

    @staticmethod
    def process_contents(path, app_ctx=None, astuple=None, asmap=None):
        """Deserializes all files within the contents directory"""
        key = os.path.basename(path)
        etype = Parsely  # ParselySchema if 'schemas' in path else Parsely
        ismappable = key in (CONFIGS_DIRNAME, SCHEMAS_DIRNAME)
        pgs = utilities.allFiles(path, entry_type=etype, app_ctx=app_ctx)
        if astuple or ismappable:
            res = {}
            for pgobj in pgs:
                configname = getattr(pgobj, 'file_name', getattr(pgobj, 'name', None))
                res[configname] = pgobj
                if hasattr(pgobj, 'abspath'): res['filepath'] = pgobj.abspath
                pass
            return utilities.tupleconverter(key, res)
        return list(pgs) if not asmap else {p.file_name: p for p in pgs}



from pyonir.parser import Parsely
from pyonir.types import IPlugin, IApp, os
from pyonir.utilities import dict_to_class

INPUT_ATTRIBUTES = tuple("accept,accept-charset,accesskey,action,align,allow,alt,async,autocapitalize,autocomplete,autofocus,autoplay,\
page.file_namebackground,bgcolor,border,buffered,capture,challenge,charset,checked,cite,class,code,codebase,color,cols,colspan,content,\
contenteditable,contextmenu,controls,coords,crossorigin,csp,data,data-*,datetime,decoding,default,defer,dir,dirname,disabled,\
download,draggable,dropzone,enctype,enterkeyhint,for,form,formaction,formenctype,formmethod,formnovalidate,formtarget,headers,\
height,hidden,high,href,hreflang,http-equiv,icon,id,importance,integrity,intrinsicsize,inputmode,ismap,itemprop,keytype,kind,\
label,lang,language,loading,list,loop,low,manifest,max,maxlength,minlength,media,min,multiple,muted,name,novalidate,open,optimum,\
pattern,ping,placeholder,poster,preload,radiogroup,readonly,referrerpolicy,rel,required,reversed,rows,rowspan,sandbox,scope,scoped,\
selected,shape,size,sizes,slot,span,\
spellcheck,src,srcdoc,srclang,srcset,start,step,style,summary,tabindex,target,title,translate,type,usemap,width,wrap".split(
    ','))


class InputTypes:
    SUBMIT = 'button'
    FIELDSET = 'fieldset'
    HTML = 'html'
    MULTISELECT = 'select-multiple'
    SELECT = 'select'
    SELECTONE = 'select'
    OPTION = 'option'
    EMAIL = 'email'
    TEXT = 'text'
    URL = 'url'
    COLOR = 'color'
    TEL = 'tel'
    PASSWORD = 'password'
    DATE = 'date'
    RADIO = 'radio'
    FILE = 'file'
    CHECKBOX = 'checkbox'
    TEXTAREA = 'textarea'
    HIDDEN = 'hidden'
    NUMBER = 'number'
    SECTION = 'section'


class ProductPrice:

    @property
    def lowest(self):
        return min(self.prices) if self.prices else None

    @staticmethod
    def process(value, base_price):
        if not value: return None

        def isInt(s):
            if isinstance(s, str):
                s = s.strip().replace('+', '')
            try:
                return "{:.2f}".format(float(s))
            except:
                return False

        price_value = isInt(value)
        upprice = None
        if price_value:
            try:
                price_value = float(price_value)
                upprice = True if isinstance(value, str) and '+' in value else False
            except:
                raise
        return price_value if not upprice else base_price + price_value

    def __init__(self, form) -> None:
        self.base_price = float(form.data.get('price', 0))
        self.prices = []
        try:
            for x in form.inputs:
                if not getattr(x, 'price_options') or not x.inputs: continue
                for y in x.inputs:
                    pr = self.process(y.props.get('value'), self.base_price)
                    if pr: self.prices.append(pr)
        except Exception as e:
            print("Form:ProductPrice parsing error found:", str(e))
            if self.base_price: self.prices = [self.base_price]
        pass


class Form(Parsely):
    parsely_extension = "form"

    @property
    def id(self):
        return self.data['form'].get('id')

    @property
    def method(self):
        return self.data['form'].get('method', 'GET')

    @property
    def action(self):
        return self.data['form'].get('action') or f"/api/v1/forms/{self.file_name}"

    @property
    def name(self):
        return self.data['form'].get('title', self.file_name)

    @property
    def type(self):
        return self.data['form'].get('type') or 'pyonir_form'

    @property
    def redirect(self):
        return self.data['form'].get('redirect')

    @property
    def js(self):
        return self.data['form'].get('js')

    @property
    def inputs(self):
        inputs = self.data['form'].get('inputs', [])
        return [FormCtrl(x, self) for x in inputs] if inputs else None

    def __init__(self, filepath: str, app_ctx: list):
        super().__init__(abspth=filepath, app_ctx=app_ctx)
        if self.type == 'product':
            self.price = ProductPrice(self)
            self.currency = "USD"


class FormCtrl:

    @property
    def price_options(self):
        """Indentifies if ctrl option has price options for product forms"""
        is_form = getattr(self.ctrl_parent, 'is_form')
        is_price_option = (not is_form and getattr(self.ctrl_parent, 'price_options')) or self._ctrl_data.get(
            'price_options', None)
        return True if is_price_option else False

    @property
    def price_increase(self):
        pval = self.props.get('value')
        return pval.startswith('+') if isinstance(pval, str) else None

    @property
    def label(self):
        return (self._ctrl_data.get('label') or '').replace('*', '')

    @property
    def required(self):
        return (self._ctrl_data.get('label') or '').startswith('*')

    @property
    def show_label(self):
        return self._ctrl_data.get('show_label') or True

    @property
    def selected(self):
        return self._ctrl_data.get('selected')

    @property
    def row_class(self):
        return self._ctrl_data.get('row_class')

    @property
    def html(self):
        return self._ctrl_data.get('html')

    @property
    def type(self):
        return self._ctrl_data.get('type') or InputTypes.HIDDEN

    @property
    def props(self):
        props = self._ctrl_data.get('props') or {}
        for prop in ['id', 'name', 'type', 'value', 'disabled', 'class']:
            v = self._ctrl_data.get(prop)
            props[prop] = v
            if prop == 'name' and not v:
                name = self.label.lower().replace(' ', '_')
                change_label = not name.startswith('item.attr')
                if self.ctrl_parent.type == 'product' and change_label:
                    name = 'item.attr.' + name
                props[prop] = name
        # self.props = props
        return props

    @property
    def inputs(self):
        if not self._ctrl_data.get('inputs'): return None
        return [FormCtrl(cinput, self) for cinput in self._ctrl_data.get('inputs')]

    def get_label_for(self):
        try:
            nm = self.ctrl_parent.label_for
            label_for = f"{nm}_{self.label}".lower().replace(' ', '_')
            return label_for
        except Exception as e:
            return str(e)

    def __init__(self, ctrl_data: dict, ctrl_parent=None):
        # assert isinstance(ctrl_data, dict), f"Expected a Dict but got a string value {ctrl_data} "
        # Normalize scalar ctrl options into dictionary types
        par_type = ctrl_parent and getattr(ctrl_parent, 'type', 'texty')
        child_type = 'option' if par_type == 'select' else par_type
        cpl = getattr(ctrl_parent, 'label_for', '')
        self._ctrl_data = ctrl_data if isinstance(ctrl_data, dict) else {"label": ctrl_data, "value": ctrl_data,
                                                                         "type": child_type}
        if not self._ctrl_data.get('label') and self.type != 'html':
            try:
                k, v = list(self._ctrl_data.items())[0]
                self._ctrl_data = {"label": k, "value": v, "type": child_type}
            except Exception as e:
                print(__name__, str(e))
        self.ctrl_parent = dict_to_class({
            "is_form": isinstance(ctrl_parent, Form),
            "label_for": cpl,
            "type": getattr(ctrl_parent, "type", None),
            "price_options": getattr(ctrl_parent, "price_options", None)
        }, 'CtrlParent')

        self.label_for = self.get_label_for()


class Forms(IPlugin):
    name = "Forms plugin"

    def __init__(self, app: IApp):
        self.forms = self.collect_dir_files(os.path.join(app.contents_dirpath, 'forms'),
                               app_ctx=app.files_ctx, file_type=Form)
        self.form_templates_dirpath = {os.path.join(os.path.dirname(__file__), 'templates')}
        self.register_templates(self.form_templates_dirpath, app)
        pass

    async def on_request(self, pyonir_req, app: IApp):
        parsely_file = pyonir_req.file
        if not isinstance(parsely_file.form, str): return
        f = getattr(self.forms, parsely_file.form)
        f.apply_filters()
        parsely_file.data['form'] = f
        pass


import os, itertools
import typing
from typing import Generator


def parse_all(path, app_ctx=None, file_type=None) -> object:
    from pyonir.parser import Parsely
    """Deserializes all files within the contents directory"""
    key = os.path.basename(path)
    etype = file_type or Parsely
    all_files_in_dir = allFiles(path, entry_type=etype, app_ctx=app_ctx)
    # Dynamically create a new class
    cls = type(key, (object,), {})

    # Create an instance of the class
    fileMap = cls()
    for pfile in all_files_in_dir:
        configname = getattr(pfile, 'file_name', getattr(pfile, 'name', None))
        setattr(fileMap, configname, pfile)
        pass
    return fileMap


def dict_to_class(data: dict, name: str):
    """
    Converts a dictionary into a class object with the given name.

    Args:
        data (dict): The dictionary to convert.
        name (str): The name of the class.

    Returns:
        object: An instance of the dynamically created class with attributes from the dictionary.
    """
    # Dynamically create a new class
    cls = type(name, (object,), {})

    # Create an instance of the class
    instance = cls()

    # Assign dictionary keys as attributes of the instance
    for key, value in data.items():
        setattr(instance, key, value)

    return instance


def get_attr(rowObj, attrPath=None, default=None, rtn_none=True):
    """
        @default returns a default obj when the target attr value is None
        @rtn_none returns a None obj when target attr value is not discovered
        otherwise the rowObj will be returned
        """
    if attrPath == None: return rowObj
    attrPath = attrPath if isinstance(attrPath, list) else attrPath.split('.')
    targetObj = None
    for key in attrPath:
        try:
            if targetObj:
                targetObj = targetObj[key]
            else:
                targetObj = rowObj.get(key)
            pass
        except Exception as TypeError:
            if targetObj:
                targetObj = getattr(targetObj, key, None)
            else:
                targetObj = getattr(rowObj, key, None)
            pass
    if (rtn_none and targetObj is None) or (default and targetObj is None):
        return default
    if not default and targetObj is None:
        return rowObj
    return targetObj


def remove_html_tags(text):
    """Remove html tags from a excerpt string"""
    import re
    clean = re.compile('<.*?>')
    return re.sub(clean, '', text)


def camel_to_snake(camel_str):
    """Converts camelcase into snake case. Thanks Chat GPT"""
    import re
    snake_str = re.sub(r'(?<!^)(?=[A-Z])', '_', camel_str).lower()
    return snake_str


def deserialize_datestr(datestr, timestr="00:00", fmt="%Y-%m-%d %H:%M:%S %p", zone="US/Eastern", auto_correct=True):
    from datetime import datetime
    import pytz
    if not isinstance(datestr, str): return datestr

    def correct_format(date_str):
        try:
            date_str = date_str.strip().replace('/', '-')
            date_str, _, timestr = date_str.partition(" ")
            timestr = timestr if timestr != "" else '12:13:14 AM'
            has_period = timestr.endswith("M")
            if not has_period:
                timestr += " AM"
            y, m, *d = date_str.split("-")
            d = "".join(d)
            fdate = f"{y}-{int(m):02d}-{d}"
            if int(y) < int(d):
                fdate = f"{d}-{int(y):02d}-{m}"
                print(f"\tIncorrect format on date string {date_str}. it should be {fdate}")
                return fdate
            return f"{fdate} {timestr}"
        except Exception as e:
            return None

    try:
        return pytz.utc.localize(datetime.strptime(datestr, fmt))
    except ValueError as e:
        # return str(e)
        if not auto_correct: return str(e)
        datestr = correct_format(datestr)
        return pytz.utc.localize(datetime.strptime(datestr, fmt)) if datestr else None


def sortBykey(listobj, sort_by_key="", limit="", reverse=True):
    """Sorts list of obj by key"""

    def get_path_object(rowObj, path):
        targetObj = None
        for key in path.split('.'):
            try:
                if targetObj:
                    targetObj = targetObj[key]
                else:
                    targetObj = rowObj[key]
                pass
            except Exception as error:
                raise error
        return targetObj

    try:
        sorted_dict = sorted(getattr(listobj, 'data', listobj), key=lambda obj: get_path_object(obj, sort_by_key),
                             reverse=reverse)
        # sorted_dict = sorted(getattr(listobj,'data', listobj), key = lambda x:x[sort_by_key], reverse=reverse)
        if limit:
            return sorted_dict[:limit]
        return sorted_dict
    except Exception as e:
        return listobj


def allFiles(abs_dirpath: str,
             app_ctx: 'IApp' = None,
             entry_type: any = None,
             rtn_attr: str = None,
             exclude_dirs: list[str] = None,
             force_all: bool = True) -> Generator:
    """Returns a generator of files from a directory path"""
    from .parser import ParselyMedia, ParselyPage
    from .parser import ALLOWED_CONTENT_EXTENSIONS, IGNORE_FILES
    if abs_dirpath in (exclude_dirs or []): return []
    if not entry_type: entry_type = ParselyPage

    _, __, base_dir, static_dir = app_ctx

    def get_datatype(parentdir, rel_filepath, etype):
        filepath = os.path.normpath(os.path.join(parentdir, rel_filepath))
        if entry_type == 'path':
            return filepath
        ispg = filepath.endswith(ALLOWED_CONTENT_EXTENSIONS)
        isschema = hasattr(etype, 'from_path')
        p = etype.from_path(filepath, app_ctx) if isschema else etype(filepath, app_ctx) \
            if ispg else ParselyMedia(filepath, app_ctx)
        return Collection.get_attr(p, rtn_attr, None)

    def is_public(parentdir, entry=None):
        if force_all: return True
        parentdir = parentdir.replace(base_dir, "").lstrip(os.path.sep)
        is_hidden_dir = parentdir.startswith(IGNORE_FILES)
        if entry in IGNORE_FILES:
            return False
        if not entry:
            return False if is_hidden_dir else True
        else:
            is_hidden_file = entry.startswith(IGNORE_FILES)
            is_filetype = entry.endswith(ALLOWED_CONTENT_EXTENSIONS)
            return False if is_filetype and is_hidden_file or is_hidden_dir else True

    for parentdir, subs, files in os.walk(os.path.normpath(abs_dirpath)):
        folderRoot = parentdir.replace(base_dir, "").lstrip(os.path.sep)
        subFolderRoot = os.path.basename(folderRoot)
        skipRoot = (folderRoot in IGNORE_FILES
                    or folderRoot.startswith(IGNORE_FILES)
                    or subFolderRoot.startswith(IGNORE_FILES))
        skipSubs = subFolderRoot in exclude_dirs if exclude_dirs else 0
        if skipRoot or skipSubs: continue

        for filename in files:
            if not force_all and not filename.endswith(ALLOWED_CONTENT_EXTENSIONS): continue
            if not is_public(parentdir, filename) or filename in IGNORE_FILES: continue
            yield get_datatype(parentdir, filename, entry_type)


def delete_file(full_filepath):
    import shutil
    if os.path.isdir(full_filepath):
        shutil.rmtree(full_filepath)
        return True
    elif os.path.isfile(full_filepath):
        os.remove(full_filepath)
        return True
    return False


def create_file(file_abspath: str, data: any = None, is_json: bool = False, mode='w') -> bool:
    def write_file(file_abspath, data, is_json=False, mode='w'):
        import json
        with open(file_abspath, mode, encoding="utf-8") as f:
            if is_json:
                json.dump(data, f, indent=2, sort_keys=True, default=json_serial)
            else:
                f.write(data)

    """Creates a new file based on provided data
    Args:
        file_abspath: str = path to proposed file
        data: any = contents to write into file
        is_json: bool = strict json file
        mode: str = write mode for file w|w+|a
    Returns:
        bool: The return value if file was created successfully
    """
    if not os.path.exists(os.path.dirname(file_abspath)):
        os.makedirs(os.path.dirname(file_abspath))
    try:

        if is_json:
            file_abspath = file_abspath.replace(".md", ".json")
        write_file(file_abspath, data, is_json=is_json, mode=mode)

        return True
    except Exception as e:
        print(f"Error create_file method: {str(e)}")
        return False


def copy_assets(src: str, dst: str, purge: bool = True):
    """Copies files from a source directory into a destination directory with option to purge destination"""
    import shutil
    from shutil import ignore_patterns
    print(f"\033[92mCoping `{src}` theme assets into {dst}")
    try:
        if os.path.exists(dst) and purge:
            shutil.rmtree(dst)
        shutil.copytree(src, dst, ignore=ignore_patterns('*.pyc', 'tmp*', 'node_modules', '.*'))
    except NotADirectoryError as e:
        shutil.copyfile(src, dst)
    except Exception as e:
        raise


def tupleconverter(tuple_name, fileobj):
    """Converts dicts into tuples"""
    if isinstance(fileobj, tuple): return fileobj
    from .parser import PROTECTED_FIELD_PREFIX, Parsely
    def tupleConverter(named, v):
        named = named.replace(' ', '_').replace('-', '_')
        if isinstance(v, dict):
            return tupleconverter(named, v)
        elif isinstance(v, Parsely):
            return tupleConverter(named, v.data)
        elif isinstance(v, list):
            return [tupleConverter(named, item) for item in v]
        else:
            return v

    def sanitize(field):
        f = field.replace(' ', '_').replace('-', '_')
        fkw = 'class, for, return, global, pass, print, raise'.split(', ')
        if f in fkw:
            print(f'WARNING: reserved keyword {f} is not allowed and has changed to {f}_ ')
            f = f + '_'
        return f

    try:
        keys = [sanitize(key) for key in fileobj.keys() if not key.startswith(PROTECTED_FIELD_PREFIX)]

        from collections import namedtuple
        return namedtuple(tuple_name, keys)(
            *tuple(tupleConverter(n, v) for n, v in fileobj.items() if
                   not n.startswith(PROTECTED_FIELD_PREFIX))
        )
    except Exception as e:
        print(f"tuple_converter: {tuple_name} {str(e)}")
        return fileobj


def json_serial(obj):
    """JSON serializer for nested objects not serializable by default jsonify"""
    from datetime import datetime
    # from pycasso.utilities import Collection
    from .parser import Parsely, ParselySchema
    if isinstance(obj, (datetime)):
        return obj.isoformat()
    elif isinstance(obj, Collection):
        return list(obj.collection)
    elif isinstance(obj, (ParselySchema, Parsely)):
        return obj.data
    else:
        return None if not hasattr(obj, '__dict__') else obj.__dict__


def secure_upload_filename(filename):
    import re
    # Strip leading and trailing whitespace from the filename
    filename = filename.strip()

    # Replace spaces with underscores
    filename = filename.replace(' ', '_')

    # Remove any remaining unsafe characters using a regular expression
    # Allow only alphanumeric characters, underscores, hyphens, dots, and slashes
    filename = re.sub(r'[^a-zA-Z0-9_.-]', '', filename)

    # Ensure the filename doesn't contain multiple consecutive dots (.) or start with one
    filename = re.sub(r'\.+', '.', filename).lstrip('.')

    # Return the filename as lowercase for consistency
    return filename.lower()


class pcolors:
    RESET = '\033[0m'
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


class Collection:
    @property
    def is_bool(self):
        return isinstance(self.where, list) and len(self.where) == 1

    @property
    def count(self):
        from sortedcontainers import SortedList
        return self.collection._len if isinstance(self.collection, SortedList) else len(self.collection)

    def __init__(self, targetIterator, sortBy: str = 'date', where: list = [], reverse=False):
        from sortedcontainers import SortedList
        self.reverse = reverse
        self.sortBy = sortBy
        self.where = where
        self.collection = SortedList(targetIterator, key=lambda x: self.get_attr(x, sortBy))
        pass

    def where_condition(self, target):
        """Evaluates if a property is true or false."""
        from datetime import datetime, timedelta
        from pyonir import Site
        if not self.where: return True
        try:
            if self.is_bool:
                return self.get_attr(target, self.where) != True
            else:
                attr_path, oper, rcon = self.where
                target = target.data if not isinstance(target, dict) else target
                lvalue = self.get_attr(target, attr_path)
                # case handling for dates
                if isinstance(lvalue, datetime):
                    if isinstance(rcon, str):
                        rcon = int(rcon)
                    if oper == '>':  # future dates
                        return lvalue > rcon
                    elif oper == '<':  # past dates
                        return lvalue < rcon
                    elif oper == 'in_last':  # within days
                        current_date = Site.system().get('current_date')
                        day_margin = timedelta(days=rcon)
                        return (current_date - day_margin) <= lvalue <= current_date + day_margin
                elif isinstance(lvalue, str):
                    # Case handling for strings
                    if oper == 'in':
                        return lvalue.find(rcon) > -1
        except Exception as e:
            # raise
            return False
        pass

    @staticmethod
    def get_attr(rowObj, attrPath=None, default=None, rtn_none=True):
        """
        @default returns a default obj when the target attr value is None
        @rtn_none returns a None obj when target attr value is None
        """
        if attrPath == None: return rowObj
        attrPath = attrPath if isinstance(attrPath, list) else attrPath.split('.')
        targetObj = None
        for key in attrPath:
            try:
                if targetObj:
                    targetObj = targetObj[key]
                else:
                    targetObj = rowObj.get(key)
                pass
            except Exception as TypeError:
                if targetObj:
                    targetObj = getattr(targetObj, key, None)
                else:
                    targetObj = getattr(rowObj, key, None)
                pass
        if (rtn_none and targetObj is None) or (default and targetObj is None):
            return default
        if not default and targetObj is None:
            return rowObj
        return targetObj
        # return targetObj if not targetObj and rtn_empty else rowObj

    def groupedBy(self, keyname, where=None, targetField=None):
        """Returns dictionary groupings based on object keyname from a list"""
        from collections import defaultdict
        results = defaultdict(list)
        if where: self.where = where
        for item in self.collection:
            item = self.get_attr(item, targetField)
            groupname = self.get_attr(item, keyname)
            try:
                if not self.where_condition(item): continue
            except NameError:
                raise Exception('`item` was not found in the groupdedBy where argument')
            if isinstance(groupname, list):
                for tag in groupname:
                    results[tag.lower().strip()].append(item)
            elif groupname is not None and isinstance(groupname, str):
                results[groupname.lower().strip()].append(item)
        return results

    def paginate(self, start: int = 0, end: int = None, reverse: bool = None, where: list = None,
                 attr_path: str = None):
        from sortedcontainers import SortedList
        if where: self.where = where
        end = self.count if end is None else end
        begin = start
        stop = end
        if begin > stop or begin == self.count or not isinstance(self.collection, SortedList): return False
        for obj in self.collection.islice(start, end, reverse=reverse or self.reverse):
            if self.where and not self.where_condition(obj): continue
            obj.set_schema()
            if obj.schema: obj = obj.schema.model
            yield get_attr(obj, attr_path, rtn_none=False) if attr_path else obj


def get_module(pkg_path: str, callable_name: str) -> tuple[any, typing.Callable]:
    from importlib import util
    pkg = util.spec_from_file_location(callable_name, pkg_path).loader.load_module(callable_name)
    func = getattr(pkg, callable_name)

    return (pkg, func)


def generate_id():
    import uuid
    return str(uuid.uuid1())


class PyonirMod:
    """Pyonir module object storing path and module class to be used later during runtime"""

    def __init__(self, path):
        from importlib import util
        from pyonir import PYONIR_DIRPATH
        assert (os.path.exists(path)), "Module path doesn't exist"

        # init_file_path = path if path.endswith('.py') else os.path.join(path, '__init__.py')
        self.class_module = None
        self.package = os.path.dirname(path) \
            .replace(os.path.dirname(PYONIR_DIRPATH), "") \
            .replace(os.path.sep, ".").lstrip(".")
        self.callable_name = os.path.basename(path).split('.py')[0]
        self.import_str = f"{self.package}.{self.callable_name}.{self.callable_name}"
        self.origin = os.path.join(path, self.callable_name + ".py") if not path.endswith('.py') else path
        if not os.path.exists(self.origin): self.origin = os.path.join(path, '__init__.py')
        try:  # Importing the module from file_path expects callable_name
            self.class_module = getattr(
                util.spec_from_file_location(self.callable_name, self.origin).loader.load_module(), self.callable_name)
        except AttributeError as e:
            raise
            # self.class_module = importlib.util.spec_from_file_location(self.callable_name, file_path).loader.load_module()
        except Exception as e:
            raise
            # print(f"{__name__} {self.origin}: {str(e)}")


class DB:
    def __init__(self):
        pass

    def connect(self, dbname: str, dbuser: str, dbpass: str, dbport: int, dbhost: str):
        pass

    @staticmethod
    async def upload_documents(documents: list['RequestFile'] = None, upload_root_dirpath: str = None):
        """Uploads files sent by server request"""
        from . import IMAGE_SIZES, UPLOADS_ROUTE, ALLOWED_UPLOAD_EXTENSIONS
        from .server import RequestFile

        async def saveMedia(file: RequestFile, foldername: str = None) -> str:

            def clean_up_path(filepath):
                return filepath.replace(file.uploads_dirpath, UPLOADS_ROUTE)

            def createImagefolder(folderpath, with_thumbs=None):
                if not os.path.exists(folderpath):
                    os.makedirs(folderpath)

                '''create separate folders for smaller sizes'''
                if with_thumbs:
                    for sizename in IMAGE_SIZES:
                        if not os.path.exists(os.path.join(folderpath, sizename)) and sizename != 'original':
                            # make the path
                            os.makedirs(os.path.join(folderpath, sizename))
                return folderpath

            def resize(filedirpath, imageFile):
                from PIL import Image
                '''
                Resize each image and save to the upload path in corresponding image size and paths
                This happens after full size images are saved to the filesystem
                '''

                try:
                    for sizename, dimensions in IMAGE_SIZES.items():
                        origimage = os.path.join(filedirpath, imageFile)
                        img = Image.open(origimage)
                        imgfilesize = os.path.getsize(origimage)
                        # img = img.resize(dimensions, Image.ANTIALIAS)
                        img.thumbnail(dimensions, Image.ANTIALIAS)
                        folderpath = os.path.join(filedirpath, sizename)
                        if not os.path.exists(folderpath):
                            os.makedirs(folderpath)
                        img.save(os.path.join(filedirpath, sizename, imageFile))

                except Exception as e:
                    pass

            def secureFile(filename):
                """ @return bool if file is allowed to upload """

                def allowed_file(filename):
                    # This function checks a given string extension
                    return '.' in filename and \
                        filename.rsplit('.', 1)[1] in ALLOWED_UPLOAD_EXTENSIONS

                return secure_upload_filename(filename) if allowed_file(filename) else None

            file_ext = file.filename.split('.').pop()
            upload_folder = os.path.join("documents", foldername) if file_ext in ("pdf", "zip") else foldername
            allowedFile = secureFile(file.filename)
            is_image = file_ext not in ('pdf', 'zip')
            if allowedFile:
                upload_dir = os.path.join(file.uploads_dirpath, upload_folder)
                folderpath = upload_dir.replace(' ', '-')
                createImagefolder(folderpath, with_thumbs=is_image)
                file.filepath = os.path.join(folderpath, allowedFile)
                # save original file
                await file.save()
                # resize original file
                if is_image: resize(file.uploads_dirpath, allowedFile)
                # save base64
                return clean_up_path(file.filepath)
            return 'file named {} was not allowed'.format(file.filename)

        upload_response = {'files': [], 'status': ''}
        try:
            upload_response['files'] = [await saveMedia(doc, upload_root_dirpath) for doc in documents]
            upload_response['status'] = "successfully uploaded {} files".format(len(documents))
        except Exception as e:
            raise
        return upload_response

    @staticmethod
    def find(
            select,
            where: str = None,
            target: str = None,
            entry_type: any = None,
            allow_all=False):
        from .parser import Parsely
        if not entry_type: entry_type = Parsely
        result = None
        selected = allFiles(select, entry_type=entry_type)
        if where and target and selected:
            for file in selected:
                if Collection.get_attr(file, where) == target:
                    result = file.file_schema.map_to_schema(file)
                    # result = file.to_dict() if rtn_file_as!='file' else file
        else:
            result = list(selected)
        return result

    @staticmethod
    def get_menus(active_page=''):
        """site navigations based on all index.md files and any file within the pages directory"""
        from pyonir import Site as app
        if app is None: return None
        assert hasattr(app, 'pages_dirpath'), "Get menus 'app' parameter does not have a pages dirpath property"

        pages = {}
        subpages = {}

        def get_nav(app):
            """Gets page nav based on file contents"""
            file_list = allFiles(app.pages_dirpath, apply_schema='menu',
                                 app_ctx=app.files_ctx)  # return files using menu schema model

            for page in file_list:
                pstatus = page.get('status')
                purl = page.get('url')
                menu_group = page.get('menu')  # parent level menu item
                menu_parent = page.get('menu_parent')  # child level menu item
                if pstatus == 'hidden' or (not menu_group and not menu_parent): continue
                page['active'] = active_page == page.get('url')
                if menu_group:
                    pages[purl] = page
                elif menu_parent:
                    _ref = subpages.get(menu_parent)
                    if not _ref:
                        subpages[menu_parent] = [page]
                    else:
                        _ref.append(page)

            if subpages:
                for k, m in subpages.items():
                    pmenu = pages.get(k)
                    if not pmenu: continue
                    pmenu.update({"sub_menu": m})

        # site navigation
        get_nav(app)
        # Collect any plugins navigation
        # if hasattr(app, 'Resolvers'):
        #     for iplg_id, iplg in app.Resolvers.__dict__.items():
        #         if not hasattr(iplg, 'app_extension') or not iplg.is_enabled: continue
        #         get_nav(iplg)

        return Collection(pages.values(), sortBy='order').groupedBy('menu')

    @staticmethod
    def tags(searchStr="", model="title,url,tag,tags"):
        import re
        """Attempts to return a dictionary of tags based on file propterties"""
        try:
            # passing a path of '/' will return all files. expensive operation here
            from . import Site
            allfiles = allFiles(Site.content_dirpath, model=model, post_process=False)
            # tagerror = "Tag named `{}` was not found".format(searchStr)
            tags_collection = {}
            query_search_wordList = searchStr.lower().replace(' ', ',').split(',') if searchStr != '' else None

            def tagContains(tagname):
                """search exact match """
                if searchStr:
                    for term in query_search_wordList:
                        # print(tagname+' <---> '+term)
                        if not term in tagname:
                            return False
                    return True

            if not query_search_wordList:
                for fileObj in allfiles:
                    fileTags = fileObj.data.get('tags', fileObj.data.get('tag'))
                    if not fileTags: continue
                    for queryWord in fileTags:
                        queryWord = queryWord.strip()
                        if not tags_collection.get(queryWord):
                            tags_collection[queryWord] = [fileObj.data]
                        else:
                            tags_collection[queryWord].append(fileObj.data)
                return tags_collection
            else:
                for fileObj in allfiles:
                    fileTags = fileObj.data.get('tags', [])
                    fileTitle = fileObj.data.get('title', '').lower().split(' ')
                    fileName = fileObj.file_name.lower().replace('-', ' ').replace('_', ' ').split(' ')
                    wordList = list(itertools.chain(fileTags, fileTitle, fileName))
                    url = fileObj.data.get('url')
                    for queryWord in query_search_wordList:
                        if tags_collection.get(url): continue
                        if queryWord not in wordList: continue
                        tags_collection.update({url: fileObj.data})
                return list(tags_collection.values())
        except Exception as e:
            raise



import dataclasses
import typing, os
from enum import Enum
from jinja2 import Environment
from starlette.applications import Starlette
from starlette.requests import Request as StarletteRequest

from pyonir.utilities import parse_all, get_attr

TEXT_RES = 'text/html'
JSON_RES = 'application/json'
EVENT_RES = 'text/event-stream'

class Parsely:
    app_ctx: tuple
    abspath: str
    absdir: str
    contents_relpath: str
    file_ctx: str
    file_dir: str
    file_type: str
    file_name: str
    file_ext: str
    file_relpath: str
    file_contents: str
    file_lines: str
    file_line_count: str
    data: dict
    schema: 'schema'


@dataclasses.dataclass
class PyonirRequest:
    raw_path: str
    method: str
    path: str
    path_params: str
    url: str
    slug: str
    query_params: dict
    parts: list
    limit: int
    model: str
    is_home: bool
    form: dict
    files: list
    ip: str
    host: str
    protocol: str
    headers: dict
    args: dict
    browser: str
    type: str
    status_code: int
    auth: any
    use_endpoints: bool
    server_response = ""
    server_request: StarletteRequest
    file: Parsely | None

    @property
    def redirect(self):
        return self.form.get('redirect')

    @staticmethod
    def process_header(headers):
        nheaders = dict(headers)
        nheaders['accept'] = nheaders.get('accept', TEXT_RES).split(',', 1)[0]
        nheaders['user-agent'] = nheaders['user-agent'].split(' ').pop().split('/', 1)[0]
        return nheaders

    def request_model(self, route_params: list[str], models: dict) -> dict:
        """Maps route params into instances of param type"""
        def init_request_model(req, param, models):
            cls = get_attr(models, param, None)
            return cls(**req.form) if cls else None

        req_name = route_params[0] if route_params else ''
        kwargs = {k: get_attr(self.path_params, k, init_request_model(self, k, models))
                  for k in route_params[1:]} if route_params else {}
        args = {f"{req_name}": self, **kwargs}
        return args

    def derive_status_code(self, is_system_control: bool):
        """Create status code for web request based on a file's availability, status_code property"""
        file_code = self.file.data.get('status_code', 200)
        return file_code if self.file and self.file.file_exists or not is_system_control else 404

    def render_error(self):
        """Data output for an unknown file path for a web request"""
        return {
            "url": self.url,
            "method": self.method,
            "status": self.status_code,
            "res": self.server_response,
            "title": f"{self.path} was not found!",
            "content": f"Perhaps this page once lived but has now been archived or permanently removed."
        }


# RoutePath: str
# RouteFunction: callable
# RouteMethods: [str]
# RequestModel: any
# PyonirRoutes: ['RoutePath', 'RouteFunction', 'RouteMethods', 'RequestModel']
# PyonirEndpoints: [('RoutePath', ['PyonirRoutes'])]


class PyonirHooks(Enum):
    ON_REQUEST = 'ON_REQUEST'
    ON_PARSELY_COMPLETE = 'ON_PARSELY_COMPLETE'


class PyonirServer(Starlette):
    ws_routes = []
    sse_routes = []
    auth_routes = []
    endpoints = []
    url_map = {}
    resolvers = {}
    response_renderer: typing.Callable
    create_route: typing.Callable
    create_endpoint: typing.Callable
    serve_static: typing.Callable
    serve_redirect: typing.Callable

    def __int__(self):
        super().__init__()


class IApp:
    def __init__(self):
        self.pages_dirpath = None
        self.frontend_dirpath = None
        self.backend_dirpath = None
        self.contents_dirpath = None
        self.files_ctx = None
        self.theme_static_dirpath = None
        self.server: PyonirServer = None
        self.site_logs_dirpath = None
        self.TemplateParser: Environment = None
        self.app_nginx_conf_filepath = None
        self.app_socket_filepath = None
        self.theme_assets_dirpath = None
        self.ssg_dirpath = None
        self.uploads_dirpath = None
        self.configs = None
        self.domain = None
        self.env = None
        self.is_dev = None
        self.is_secure = None
        self.host = None
        self.port = None
        self.name = None
        self.app_dirpath = None

    pass

    def run_plugins(self, hook: PyonirHooks, pyonir_request): pass


class IPlugin:
    name: str
    hooks: PyonirHooks

    @staticmethod
    def collect_dir_files(dir_path: str, app_ctx: tuple, file_type: any) -> [Parsely]:
        return parse_all(dir_path, app_ctx, file_type)

    @staticmethod
    def uninstall(app):
        """Uninstall plugin from system. this method will destroy any traces of the plugin and its files"""
        pass

    @staticmethod
    def register_templates(dir_paths: [str], app: IApp):
        """Loads new jinja templates paths to application instance"""
        if not hasattr(app.TemplateParser, 'loader'): return None
        for path in dir_paths:
            if path in app.TemplateParser.loader.searchpath: continue
            app.TemplateParser.loader.searchpath.append(path)

    @staticmethod
    def unregister_templates(dir_paths: [str], app: IApp):
        """Loads new jinja templates paths to application instance"""
        if not hasattr(app.TemplateParser, 'loader'): return None
        for path in dir_paths:
            if path in app.TemplateParser.loader.searchpath: continue
            app.TemplateParser.loader.searchpath.remove(path)



from setuptools import setup, find_packages

with open("README.md", "r") as f:
    description = f.read()

setup(
    name='pyonir',
    description='a python library for building web applications',
    long_description=description,
    long_description_content_type="text/markdown",
    url='https://pyonir.dev',
    author='Derry Spann',
    author_email='pyonir@derryspann.com',
    version='0.0.9',
    packages=find_packages(),
    package_data={
        'pyonir': ['libs/*']
    },
    install_requires=['starlette', 'inquirer', 'uvicorn', 'starlette_session', 'starlette_wtf', 'pytz',
                      'sortedcontainers', 'jinja2', 'webassets'],
    include_package_data=True,
    entry_points={
        "console_scripts": [
            "pyonir-create = pyonir:cli.PyonirSetup"
        ]
    },
    python_requires=">=3.9"
)



import os
import pyonir
from backend.controllers import endpoints
# Instantiate pyonir application
demo_app = pyonir.init(os.path.dirname(__file__))

# Install plugins
# demo_app.register_plugins([YOUR_PLUGIN_MODULE_HERE])

# Generate static website
# demo_app.generate_static_website()

# Run server
demo_app.run(endpoints=endpoints)
